@@ -26,7 +26,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 static ulong
 to_FlxqX(GEN P, GEN Q, GEN T, GEN p, GEN *pt_P, GEN *pt_Q, GEN *pt_T)
 {
-  ulong pp = (ulong)p[2];
+  ulong pp = uel(p,2);
   long v = get_FpX_var(T);
   *pt_P = ZXX_to_FlxX(P, pp, v);
   if (pt_Q) *pt_Q = ZXX_to_FlxX(Q, pp, v);
@@ -152,11 +152,55 @@ FpXX_mulu(GEN P, ulong u, GEN p)
   return FpXX_renormalize(res,lP);
 }
 
+GEN
+FpXX_deriv(GEN P, GEN p)
+{
+  long i, l = lg(P)-1;
+  GEN res;
+
+  if (l < 3) return pol_0(varn(P));
+  res = cgetg(l, t_POL);
+  res[1] = P[1];
+  for (i=2; i<l ; i++)
+  {
+    GEN x = gel(P,i+1);
+    gel(res,i) = typ(x)==t_INT? Fp_mulu(x,i-1,p): FpX_mulu(x,i-1,p);
+  }
+  return FpXX_renormalize(res, l);
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                             (Fp[X]/(Q))[Y]                      */
 /*                                                                 */
 /*******************************************************************/
+
+static GEN
+get_FpXQX_red(GEN T, GEN *B)
+{
+  if (typ(T)!=t_VEC) { *B=NULL; return T; }
+  *B = gel(T,1); return gel(T,2);
+}
+
+GEN
+get_FpXQX_mod(GEN T) { return typ(T)==t_VEC? gel(T,2): T; }
+
+long
+get_FpXQX_var(GEN T) { return typ(T)==t_VEC? varn(gel(T,2)): varn(T); }
+
+long
+get_FpXQX_degree(GEN T) { return typ(T)==t_VEC? degpol(gel(T,2)): degpol(T); }
+
+GEN
+random_FpXQX(long d1, long v, GEN T, GEN p)
+{
+  long dT = get_FpX_degree(T), vT = get_FpX_var(T);
+  long i, d = d1+2;
+  GEN y = cgetg(d,t_POL); y[1] = evalsigne(1) | evalvarn(v);
+  for (i=2; i<d; i++) gel(y,i) = random_FpX(dT, vT, p);
+  return FpXQX_renormalize(y,d);
+}
+
 /*Not stack clean*/
 GEN
 Kronecker_to_FpXQX(GEN Z, GEN T, GEN p)
@@ -206,6 +250,10 @@ GEN
 ZXX_mul_Kronecker(GEN x, GEN y, long n)
 { return ZX_mul(ZXX_to_Kronecker(x,n), ZXX_to_Kronecker(y,n)); }
 
+GEN
+ZXX_sqr_Kronecker(GEN x, long n)
+{ return ZX_sqr(ZXX_to_Kronecker(x,n)); }
+
 GEN
 FpXQX_red(GEN z, GEN T, GEN p)
 {
@@ -255,10 +303,11 @@ static GEN
 FpXY_FpY_mulspec(GEN x, GEN y, GEN T, GEN p, long lx, long ly)
 {
   pari_sp av = avma;
-  GEN z = RgXY_swapspec(x,get_FpX_degree(T)-1,MAXVARN,lx);
-  z = FpXX_FpX_mulspec(z,y,p,MAXVARN,ly);
+  long v = fetch_var();
+  GEN z = RgXY_swapspec(x,get_FpX_degree(T)-1,v,lx);
+  z = FpXX_FpX_mulspec(z,y,p,v,ly);
   z = RgXY_swapspec(z+2,lx+ly+3,get_FpX_var(T),lgpol(z));
-  return gerepilecopy(av,z);
+  (void)delete_var(); return gerepilecopy(av,z);
 }
 
 static GEN
@@ -316,11 +365,10 @@ FpXQX_FpXQ_mul(GEN P, GEN U, GEN T, GEN p)
 static GEN
 FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
 {
-  long vx, dx, dy, dz, i, j, sx, lr;
+  long vx, dx, dy, dy1, dz, i, j, sx, lr;
   pari_sp av0, av, tetpil;
   GEN z,p1,rem,lead;
 
-  if (!T) return FpX_divrem(x,y,p,pr);
   if (!signe(y)) pari_err_INV("FpX_divrem",y);
   vx=varn(x); dy=degpol(y); dx=degpol(x);
   if (dx < dy)
@@ -334,7 +382,7 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
     }
     return pol_0(vx);
   }
-  lead = leading_term(y);
+  lead = leading_coeff(y);
   if (!dy) /* y is constant */
   {
     if (pr && pr != ONLY_DIVIDES)
@@ -347,34 +395,18 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
     return gerepileupto(av0,x);
   }
   av0 = avma; dz = dx-dy;
-  if (lgefint(p) == 3)
-  { /* assume ab != 0 mod p */
-    {
-      GEN *gptr[2];
-      GEN a, b, t;
-      ulong pp = to_FlxqX(x, y, T, p, &a, &b, &t);
-      z = FlxqX_divrem(a,b,t,pp,pr);
-      tetpil=avma;
-      z = FlxX_to_ZXX(z);
-      if (pr && pr != ONLY_DIVIDES && pr != ONLY_REM)
-        *pr = FlxX_to_ZXX(*pr);
-      else return gerepile(av0,tetpil,z);
-      gptr[0]=pr; gptr[1]=&z;
-      gerepilemanysp(av0,tetpil,gptr,2);
-      return z;
-    }
-  }
   lead = gequal1(lead)? NULL: gclone(Fq_inv(lead,T,p));
   avma = av0;
   z = cgetg(dz+3,t_POL); z[1] = x[1];
   x += 2; y += 2; z += 2;
+  for (dy1=dy-1; dy1>=0 && !signe(gel(y, dy1)); dy1--);
 
   p1 = gel(x,dx); av = avma;
   gel(z,dz) = lead? gerepileupto(av, Fq_mul(p1,lead, T, p)): gcopy(p1);
   for (i=dx-1; i>=dy; i--)
   {
     av=avma; p1=gel(x,i);
-    for (j=i-dy+1; j<=i && j<=dz; j++)
+    for (j=i-dy1; j<=i && j<=dz; j++)
       p1 = Fq_sub(p1, Fq_mul(gel(z,j),gel(y,i-j),NULL,p),NULL,p);
     if (lead) p1 = Fq_mul(p1, lead, NULL,p);
     tetpil=avma; gel(z,i-dy) = gerepile(av,tetpil,Fq_red(p1,T,p));
@@ -385,7 +417,7 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
   for (sx=0; ; i--)
   {
     p1 = gel(x,i);
-    for (j=0; j<=i && j<=dz; j++)
+    for (j=maxss(0,i-dy1); j<=i && j<=dz; j++)
       p1 = Fq_sub(p1, Fq_mul(gel(z,j),gel(y,i-j),NULL,p),NULL,p);
     tetpil=avma; p1 = Fq_red(p1, T, p); if (signe(p1)) { sx = 1; break; }
     if (!i) break;
@@ -405,7 +437,7 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
   for (i--; i>=0; i--)
   {
     av=avma; p1 = gel(x,i);
-    for (j=0; j<=i && j<=dz; j++)
+    for (j=maxss(0,i-dy1); j<=i && j<=dz; j++)
       p1 = Fq_sub(p1, Fq_mul(gel(z,j),gel(y,i-j), NULL,p), NULL,p);
     tetpil=avma; gel(rem,i) = gerepile(av,tetpil, Fq_red(p1, T, p));
   }
@@ -416,25 +448,229 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
   *pr = rem; return z-2;
 }
 
-GEN
-FpXQX_gcd(GEN P, GEN Q, GEN T, GEN p)
+static GEN
+FpXQX_halfgcd_basecase(GEN a, GEN b, GEN T, GEN p)
 {
-  pari_sp av=avma, av0;
-  GEN R;
+  pari_sp av=avma;
+  GEN u,u1,v,v1;
+  long vx = varn(a);
+  long n = lgpol(a)>>1;
+  u1 = v = pol_0(vx);
+  u = v1 = pol_1(vx);
+  while (lgpol(b)>n)
+  {
+    GEN r, q = FpXQX_divrem(a,b, T, p, &r);
+    a = b; b = r; swap(u,u1); swap(v,v1);
+    u1 = FpXX_sub(u1, FpXQX_mul(u, q, T, p), p);
+    v1 = FpXX_sub(v1, FpXQX_mul(v, q ,T, p), p);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpXQX_halfgcd (d = %ld)",degpol(b));
+      gerepileall(av,6, &a,&b,&u1,&v1,&u,&v);
+    }
+  }
+  return gerepilecopy(av, mkmat2(mkcol2(u,u1), mkcol2(v,v1)));
+}
+static GEN
+FpXQX_addmulmul(GEN u, GEN v, GEN x, GEN y, GEN T, GEN p)
+{
+  return FpXX_add(FpXQX_mul(u, x, T, p),FpXQX_mul(v, y, T, p), p);
+}
+
+static GEN
+FpXQXM_FpXQX_mul2(GEN M, GEN x, GEN y, GEN T, GEN p)
+{
+  GEN res = cgetg(3, t_COL);
+  gel(res, 1) = FpXQX_addmulmul(gcoeff(M,1,1), gcoeff(M,1,2), x, y, T, p);
+  gel(res, 2) = FpXQX_addmulmul(gcoeff(M,2,1), gcoeff(M,2,2), x, y, T, p);
+  return res;
+}
+
+static GEN
+FpXQXM_mul2(GEN A, GEN B, GEN T, GEN p)
+{
+  GEN A11=gcoeff(A,1,1),A12=gcoeff(A,1,2), B11=gcoeff(B,1,1),B12=gcoeff(B,1,2);
+  GEN A21=gcoeff(A,2,1),A22=gcoeff(A,2,2), B21=gcoeff(B,2,1),B22=gcoeff(B,2,2);
+  GEN M1 = FpXQX_mul(FpXX_add(A11,A22, p), FpXX_add(B11,B22, p), T, p);
+  GEN M2 = FpXQX_mul(FpXX_add(A21,A22, p), B11, T, p);
+  GEN M3 = FpXQX_mul(A11, FpXX_sub(B12,B22, p), T, p);
+  GEN M4 = FpXQX_mul(A22, FpXX_sub(B21,B11, p), T, p);
+  GEN M5 = FpXQX_mul(FpXX_add(A11,A12, p), B22, T, p);
+  GEN M6 = FpXQX_mul(FpXX_sub(A21,A11, p), FpXX_add(B11,B12, p), T, p);
+  GEN M7 = FpXQX_mul(FpXX_sub(A12,A22, p), FpXX_add(B21,B22, p), T, p);
+  GEN T1 = FpXX_add(M1,M4, p), T2 = FpXX_sub(M7,M5, p);
+  GEN T3 = FpXX_sub(M1,M2, p), T4 = FpXX_add(M3,M6, p);
+  retmkmat2(mkcol2(FpXX_add(T1,T2, p), FpXX_add(M2,M4, p)),
+            mkcol2(FpXX_add(M3,M5, p), FpXX_add(T3,T4, p)));
+}
+/* Return [0,1;1,-q]*M */
+static GEN
+FpXQX_FpXQXM_qmul(GEN q, GEN M, GEN T, GEN p)
+{
+  GEN u, v, res = cgetg(3, t_MAT);
+  u = FpXX_sub(gcoeff(M,1,1), FpXQX_mul(gcoeff(M,2,1), q, T, p), p);
+  gel(res,1) = mkcol2(gcoeff(M,2,1), u);
+  v = FpXX_sub(gcoeff(M,1,2), FpXQX_mul(gcoeff(M,2,2), q, T, p), p);
+  gel(res,2) = mkcol2(gcoeff(M,2,2), v);
+  return res;
+}
+
+static GEN
+matid2_FpXQXM(long v)
+{
+  retmkmat2(mkcol2(pol_1(v),pol_0(v)),
+            mkcol2(pol_0(v),pol_1(v)));
+}
+
+static GEN
+FpXQX_halfgcd_split(GEN x, GEN y, GEN T, GEN p)
+{
+  pari_sp av=avma;
+  GEN R, S, V;
+  GEN y1, r, q;
+  long l = lgpol(x), n = l>>1, k;
+  if (lgpol(y)<=n) return matid2_FpXQXM(varn(x));
+  R = FpXQX_halfgcd(RgX_shift_shallow(x,-n),RgX_shift_shallow(y,-n), T, p);
+  V = FpXQXM_FpXQX_mul2(R,x,y, T, p); y1 = gel(V,2);
+  if (lgpol(y1)<=n) return gerepilecopy(av, R);
+  q = FpXQX_divrem(gel(V,1), y1, T, p, &r);
+  k = 2*n-degpol(y1);
+  S = FpXQX_halfgcd(RgX_shift_shallow(y1,-k), RgX_shift_shallow(r,-k), T, p);
+  return gerepileupto(av, FpXQXM_mul2(S,FpXQX_FpXQXM_qmul(q,R, T, p), T, p));
+}
+
+/* Return M in GL_2(Fp[X]) such that:
+if [a',b']~=M*[a,b]~ then degpol(a')>= (lgpol(a)>>1) >degpol(b')
+*/
+
+static GEN
+FpXQX_halfgcd_i(GEN x, GEN y, GEN T, GEN p)
+{
+  if (lg(x)<=FpXQX_HALFGCD_LIMIT) return FpXQX_halfgcd_basecase(x, y, T, p);
+  return FpXQX_halfgcd_split(x, y, T, p);
+}
+
+GEN
+FpXQX_halfgcd(GEN x, GEN y, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  GEN M,q,r;
+  if (lgefint(p)==3)
+  {
+    ulong pp = to_FlxqX(x, y, T, p, &x, &y, &T);
+    M = FlxXM_to_ZXXM(FlxqX_halfgcd(x, y, T, pp));
+  }
+  else
+  {
+    if (!signe(x))
+    {
+      long v = varn(x);
+      retmkmat2(mkcol2(pol_0(v),pol_1(v)),
+                mkcol2(pol_1(v),pol_0(v)));
+    }
+    if (degpol(y)<degpol(x)) return FpXQX_halfgcd_i(x, y, T, p);
+    q = FpXQX_divrem(y, x, T, p, &r);
+    M = FpXQX_halfgcd_i(x, r, T, p);
+    gcoeff(M,1,1) = FpXX_sub(gcoeff(M,1,1), FpXQX_mul(q, gcoeff(M,1,2), T, p), p);
+    gcoeff(M,2,1) = FpXX_sub(gcoeff(M,2,1), FpXQX_mul(q, gcoeff(M,2,2), T, p), p);
+  }
+  return gerepilecopy(av, M);
+}
+
+static GEN
+FpXQX_gcd_basecase(GEN a, GEN b, GEN T, GEN p)
+{
+  pari_sp av = avma, av0=avma;
+  while (signe(b))
+  {
+    GEN c;
+    if (gc_needed(av0,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpXQX_gcd (d = %ld)",degpol(b));
+      gerepileall(av0,2, &a,&b);
+    }
+    av = avma; c = FpXQX_rem(a, b, T, p); a=b; b=c;
+  }
+  avma = av; return a;
+}
+
+GEN
+FpXQX_gcd(GEN x, GEN y, GEN T, GEN p)
+{
+  pari_sp av = avma;
   if (lgefint(p) == 3)
   {
     GEN Pl, Ql, Tl, U;
-    ulong pp = to_FlxqX(P, Q, T, p, &Pl, &Ql, &Tl);
+    ulong pp = to_FlxqX(x, y, T, p, &Pl, &Ql, &Tl);
     U  = FlxqX_gcd(Pl, Ql, Tl, pp);
     return gerepileupto(av, FlxX_to_ZXX(U));
   }
-  P = FpXX_red(P, p); av0 = avma;
-  Q = FpXX_red(Q, p);
-  while (signe(Q))
+  x = FpXQX_red(x, T, p);
+  y = FpXQX_red(y, T, p);
+  if (!signe(x)) return gerepileupto(av, y);
+  while (lg(y)>FpXQX_GCD_LIMIT)
   {
-    av0 = avma; R = FpXQX_rem(P,Q,T,p); P=Q; Q=R;
+    GEN c;
+    if (lgpol(y)<=(lgpol(x)>>1))
+    {
+      GEN r = FpXQX_rem(x, y, T, p);
+      x = y; y = r;
+    }
+    c = FpXQXM_FpXQX_mul2(FpXQX_halfgcd(x,y, T, p), x, y, T, p);
+    x = gel(c,1); y = gel(c,2);
+    gerepileall(av,2,&x,&y);
   }
-  avma = av0; return gerepileupto(av, P);
+  return gerepileupto(av, FpXQX_gcd_basecase(x, y, T, p));
+}
+
+static GEN
+FpXQX_extgcd_basecase(GEN a, GEN b, GEN T, GEN p, GEN *ptu, GEN *ptv)
+{
+  pari_sp av=avma;
+  GEN u,v,d,d1,v1;
+  long vx = varn(a);
+  d = a; d1 = b;
+  v = pol_0(vx); v1 = pol_1(vx);
+  while (signe(d1))
+  {
+    GEN r, q = FpXQX_divrem(d, d1, T, p, &r);
+    v = FpXX_sub(v,FpXQX_mul(q,v1,T, p),p);
+    u=v; v=v1; v1=u;
+    u=r; d=d1; d1=u;
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpXQX_extgcd (d = %ld)",degpol(d));
+      gerepileall(av,5, &d,&d1,&u,&v,&v1);
+    }
+  }
+  if (ptu) *ptu = FpXQX_div(FpXX_sub(d,FpXQX_mul(b,v, T, p), p), a, T, p);
+  *ptv = v; return d;
+}
+
+static GEN
+FpXQX_extgcd_halfgcd(GEN x, GEN y, GEN T, GEN p, GEN *ptu, GEN *ptv)
+{
+  pari_sp av=avma;
+  GEN u,v,R = matid2_FpXQXM(varn(x));
+  while (lg(y)>FpXQX_EXTGCD_LIMIT)
+  {
+    GEN M, c;
+    if (lgpol(y)<=(lgpol(x)>>1))
+    {
+      GEN r, q = FpXQX_divrem(x, y, T, p, &r);
+      x = y; y = r;
+      R = FpXQX_FpXQXM_qmul(q, R, T, p);
+    }
+    M = FpXQX_halfgcd(x,y, T, p);
+    c = FpXQXM_FpXQX_mul2(M, x,y, T, p);
+    R = FpXQXM_mul2(M, R, T, p);
+    x = gel(c,1); y = gel(c,2);
+    gerepileall(av,3,&x,&y,&R);
+  }
+  y = FpXQX_extgcd_basecase(x,y, T, p, &u,&v);
+  if (ptu) *ptu = FpXQX_addmulmul(u,v,gcoeff(R,1,1),gcoeff(R,2,1), T, p);
+  *ptv = FpXQX_addmulmul(u,v,gcoeff(R,1,2),gcoeff(R,2,2), T, p);
+  return y;
 }
 
 /* x and y in Z[Y][X], return lift(gcd(x mod T,p, y mod T,p)). Set u and v st
@@ -442,8 +678,7 @@ FpXQX_gcd(GEN P, GEN Q, GEN T, GEN p)
 GEN
 FpXQX_extgcd(GEN x, GEN y, GEN T, GEN p, GEN *ptu, GEN *ptv)
 {
-  GEN a, b, q, r, u, v, d, d1, v1;
-  long vx = varn(x);
+  GEN d;
   pari_sp ltop=avma;
   if (lgefint(p) == 3)
   {
@@ -456,18 +691,12 @@ FpXQX_extgcd(GEN x, GEN y, GEN T, GEN p, GEN *ptu, GEN *ptv)
   }
   else
   {
-    a = FpXQX_red(x, T, p);
-    b = FpXQX_red(y, T, p);
-    d = a; d1 = b; v = pol_0(vx); v1 = pol_1(vx);
-    while (signe(d1))
-    {
-      q = FpXQX_divrem(d,d1,T,p, &r);
-      v = FpXX_sub(v, FpXQX_mul(q,v1, T,p), p);
-      u=v; v=v1; v1=u;
-      u=r; d=d1; d1=u;
-    }
-    if (ptu) *ptu = FpXQX_div(FpXX_sub(d, FpXQX_mul(b,v, T,p), p),a, T,p);
-    *ptv = v;
+    x = FpXQX_red(x, T, p);
+    y = FpXQX_red(y, T, p);
+    if (lg(y)>FpXQX_EXTGCD_LIMIT)
+      d = FpXQX_extgcd_halfgcd(x, y, T, p, ptu, ptv);
+    else
+      d = FpXQX_extgcd_basecase(x, y, T, p, ptu, ptv);
   }
   gerepileall(ltop,ptu?3:2,&d,ptv,ptu);
   return d;
@@ -566,10 +795,6 @@ FpXQX_invBarrett_Newton(GEN S, GEN T, GEN p)
   return gerepilecopy(av, x);
 }
 
-const long FpXQX_INVBARRETT_LIMIT = 50;
-const long FpXQX_DIVREM_BARRETT_LIMIT = 50;
-const long FpXQX_REM_BARRETT_LIMIT = 50;
-
 /* 1/polrecip(S)+O(x^(deg(S)-1)) */
 GEN
 FpXQX_invBarrett(GEN S, GEN T, GEN p)
@@ -595,6 +820,14 @@ FpXQX_invBarrett(GEN S, GEN T, GEN p)
   return gerepileupto(ltop, r);
 }
 
+GEN
+FpXQX_get_red(GEN S, GEN T, GEN p)
+{
+  if (typ(S)==t_POL && lg(S)>FpXQX_BARRETT_LIMIT)
+    retmkvec2(FpXQX_invBarrett(S,T,p),S);
+  return S;
+}
+
 /* Compute x mod S where 2 <= degpol(S) <= l+1 <= 2*(degpol(S)-1)
  * and mg is the Barrett inverse of S. */
 static GEN
@@ -681,27 +914,32 @@ FpXQX_divrem_Barrett_noGC(GEN x, GEN mg, GEN S, GEN T, GEN p, GEN *pr)
 }
 
 GEN
-FpXQX_divrem_Barrett(GEN x, GEN B, GEN S, GEN T, GEN p, GEN *pr)
-{
-  pari_sp av = avma;
-  GEN q = FpXQX_divrem_Barrett_noGC(x,B,S,T,p,pr);
-  if (!q) {avma=av; return NULL;}
-  if (!pr || pr==ONLY_REM || pr==ONLY_DIVIDES) return gerepilecopy(av, q);
-  gerepileall(av,2,&q,pr);
-  return q;
-}
-
-GEN
-FpXQX_divrem(GEN x, GEN y, GEN T, GEN p, GEN *pr)
+FpXQX_divrem(GEN x, GEN S, GEN T, GEN p, GEN *pr)
 {
+  GEN B, y = get_FpXQX_red(S, &B);
   long dy = degpol(y), dx = degpol(x), d = dx-dy;
   if (pr==ONLY_REM) return FpXQX_rem(x, y, T, p);
-  if (d+3 < FpXQX_DIVREM_BARRETT_LIMIT)
+  if (lgefint(p) == 3)
+  {
+    GEN a, b, t, z;
+    pari_sp tetpil, av = avma;
+    ulong pp = to_FlxqX(x, y, T, p, &a, &b, &t);
+    z = FlxqX_divrem(a, b, t, pp, pr);
+    if (pr == ONLY_DIVIDES && !z) { avma = av; return NULL; }
+    tetpil=avma;
+    z = FlxX_to_ZXX(z);
+    if (pr && pr != ONLY_DIVIDES && pr != ONLY_REM)
+      *pr = FlxX_to_ZXX(*pr);
+    else return gerepile(av, tetpil, z);
+    gerepileallsp(av,tetpil,2, pr, &z);
+    return z;
+  }
+  if (!B && d+3 < FpXQX_DIVREM_BARRETT_LIMIT)
     return FpXQX_divrem_basecase(x,y,T,p,pr);
   else
   {
     pari_sp av=avma;
-    GEN mg = FpXQX_invBarrett(y, T, p);
+    GEN mg = B? B: FpXQX_invBarrett(y, T, p);
     GEN q = FpXQX_divrem_Barrett_noGC(x,mg,y,T,p,pr);
     if (!q) {avma=av; return NULL;}
     if (!pr || pr==ONLY_DIVIDES) return gerepilecopy(av, q);
@@ -711,34 +949,81 @@ FpXQX_divrem(GEN x, GEN y, GEN T, GEN p, GEN *pr)
 }
 
 GEN
-FpXQX_rem_Barrett(GEN x, GEN mg, GEN S, GEN T, GEN p)
-{
-  pari_sp av = avma;
-  return gerepileupto(av, FpXQX_divrem_Barrett_noGC(x,mg,S, T, p, ONLY_REM));
-}
-
-GEN
-FpXQX_rem(GEN x, GEN y, GEN T, GEN p)
+FpXQX_rem(GEN x, GEN S, GEN T, GEN p)
 {
+  GEN B, y = get_FpXQX_red(S, &B);
   long dy = degpol(y), dx = degpol(x), d = dx-dy;
   if (d < 0) return FpXQX_red(x, T, p);
-  if (d+3 < FpXQX_REM_BARRETT_LIMIT)
+  if (lgefint(p) == 3)
+  {
+    pari_sp av = avma;
+    GEN a, b, t, z;
+    ulong pp = to_FlxqX(x, y, T, p, &a, &b, &t);
+    z = FlxqX_rem(a, b, t, pp);
+    z = FlxX_to_ZXX(z);
+    return gerepileupto(av, z);
+  }
+  if (!B && d+3 < FpXQX_REM_BARRETT_LIMIT)
     return FpXQX_divrem_basecase(x,y, T, p, ONLY_REM);
   else
   {
     pari_sp av=avma;
-    GEN mg = FpXQX_invBarrett(y, T, p);
+    GEN mg = B? B: FpXQX_invBarrett(y, T, p);
     GEN r = FpXQX_divrem_Barrett_noGC(x, mg, y, T, p, ONLY_REM);
     return gerepileupto(av, r);
   }
 }
 
-struct _FpXQX { GEN T,p; };
+/* x + y*z mod p */
+INLINE GEN
+Fq_addmul(GEN x, GEN y, GEN z, GEN T, GEN p)
+{
+  pari_sp av;
+  if (!signe(y) || !signe(z)) return Fq_red(x, T, p);
+  if (!signe(x)) return Fq_mul(z,y, T, p);
+  av = avma;
+  return gerepileupto(av, Fq_add(x, Fq_mul(y, z, T, p), T, p));
+}
+
+GEN
+FpXQX_div_by_X_x(GEN a, GEN x, GEN T, GEN p, GEN *r)
+{
+  long l = lg(a)-1, i;
+  GEN z = cgetg(l, t_POL);
+  z[1] = evalsigne(1) | evalvarn(0);
+  gel(z, l-1) = gel(a,l);
+  for (i=l-2; i>1; i--) /* z[i] = a[i+1] + x*z[i+1] */
+    gel(z, i) = Fq_addmul(gel(a,i+1), x, gel(z,i+1), T, p);
+  if (r) *r = Fq_addmul(gel(a,2), x, gel(z,2), T, p);
+  return z;
+}
+
+struct _FpXQXQ {
+  GEN T, S;
+  GEN p;
+};
+
 static GEN _FpXQX_mul(void *data, GEN a,GEN b)
 {
-  struct _FpXQX *d=(struct _FpXQX*)data;
+  struct _FpXQXQ *d=(struct _FpXQXQ*)data;
   return FpXQX_mul(a,b,d->T,d->p);
 }
+
+static GEN _FpXQX_sqr(void *data, GEN a)
+{
+  struct _FpXQXQ *d=(struct _FpXQXQ*)data;
+  return FpXQX_sqr(a, d->T, d->p);
+}
+
+GEN
+FpXQX_powu(GEN x, ulong n, GEN T, GEN p)
+{
+  struct _FpXQXQ D;
+  if (n==0) return pol_1(varn(x));
+  D.T = T; D.p = p;
+  return gen_powu(x, n, (void *)&D, _FpXQX_sqr, _FpXQX_mul);
+}
+
 GEN
 FpXQXV_prod(GEN V, GEN T, GEN p)
 {
@@ -753,13 +1038,57 @@ FpXQXV_prod(GEN V, GEN T, GEN p)
   }
   else
   {
-    struct _FpXQX d;
-    d.p=p;
-    d.T=T;
-    return divide_conquer_assoc(V, (void*)&d, &_FpXQX_mul);
+    struct _FpXQXQ d;
+    d.T=T; d.p=p;
+    return gen_product(V, (void*)&d, &_FpXQX_mul);
   }
 }
 
+static GEN
+_FpXQX_divrem(void * E, GEN x, GEN y, GEN *r)
+{
+  struct _FpXQXQ *d = (struct _FpXQXQ *) E;
+  return FpXQX_divrem(x, y, d->T, d->p, r);
+}
+
+static GEN
+_FpXQX_add(void * E, GEN x, GEN y)
+{
+  struct _FpXQXQ *d = (struct _FpXQXQ *) E;
+  return FpXX_add(x, y, d->p);
+}
+
+static GEN
+_FpXQX_sub(void * E, GEN x, GEN y) {
+  struct _FpXQXQ *d = (struct _FpXQXQ*) E;
+  return FpXX_sub(x,y, d->p);
+}
+
+static struct bb_ring FpXQX_ring = { _FpXQX_add, _FpXQX_mul, _FpXQX_sqr };
+
+GEN
+FpXQX_digits(GEN x, GEN B, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  long d = degpol(B), n = (lgpol(x)+d-1)/d;
+  GEN z;
+  struct _FpXQXQ D;
+  D.T = T; D.p = p;
+  z = gen_digits(x, B, n, (void *)&D, &FpXQX_ring, _FpXQX_divrem);
+  return gerepileupto(av, z);
+}
+
+GEN
+FpXQX_fromdigits(GEN x, GEN B, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  struct _FpXQXQ D;
+  GEN z;
+  D.T = T; D.p = p;
+  z = gen_fromdigits(x,B,(void *)&D, &FpXQX_ring);
+  return gerepileupto(av, z);
+}
+
 /* Q an FpXY (t_POL with FpX coeffs), evaluate at X = x */
 GEN
 FpXY_evalx(GEN Q, GEN x, GEN p)
@@ -781,14 +1110,12 @@ FpXY_evaly(GEN Q, GEN y, GEN p, long vx)
   pari_sp av = avma;
   long i, lb = lg(Q);
   GEN z;
-  if (lb == 2) return pol_0(vx);
-  z = gel(Q, lb-1);
-  if (lb == 3) return typ(z)==t_INT? scalar_ZX(z, vx): ZX_copy(z);
-  if (!signe(y)) {
+  if (!signe(Q)) return pol_0(vx);
+  if (lb == 3 || !signe(y)) {
     z = gel(Q, 2);
     return typ(z)==t_INT? scalar_ZX(z, vx): ZX_copy(z);
   }
-
+  z = gel(Q, lb-1);
   if (typ(z) == t_INT) z = scalar_ZX_shallow(z, vx);
   for (i=lb-2; i>=2; i--) z = Fq_add(gel(Q,i), FpX_Fp_mul(z, y, p), NULL, p);
   return gerepileupto(av, z);
@@ -802,16 +1129,26 @@ FpXY_eval(GEN Q, GEN y, GEN x, GEN p)
 }
 
 GEN
-FpXY_FpXQ_evalx(GEN P, GEN x, GEN T, GEN p)
+FpXY_FpXQV_evalx(GEN P, GEN x, GEN T, GEN p)
 {
   long i, lP = lg(P);
   GEN res = cgetg(lP,t_POL);
   res[1] = P[1];
   for(i=2; i<lP; i++)
-    gel(res,i) = FpX_FpXQ_eval(gel(P,i), x, T, p);
+    gel(res,i) = typ(gel(P,i))==t_INT? icopy(gel(P,i)):
+                                       FpX_FpXQV_eval(gel(P,i), x, T, p);
   return FlxX_renormalize(res, lP);
 }
 
+GEN
+FpXY_FpXQ_evalx(GEN P, GEN x, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  long n = brent_kung_optpow(get_FpX_degree(T)-1,lgpol(P),1);
+  GEN xp = FpXQ_powers(x, n, T, p);
+  return gerepileupto(av, FpXY_FpXQV_evalx(P, xp, T, p));
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                       (Fp[X]/T(X))[Y] / S(Y)                    */
@@ -888,7 +1225,7 @@ FpXQXQ_sqr(GEN x, GEN S, GEN T, GEN p) {
 GEN
 FpXQXQ_invsafe(GEN x, GEN S, GEN T, GEN p)
 {
-  GEN V, z = FpXQX_extgcd(S, x, T, p, NULL, &V);
+  GEN V, z = FpXQX_extgcd(get_FpXQX_mod(S), x, T, p, NULL, &V);
   if (degpol(z)) return NULL;
   z = gel(z,2);
   z = typ(z)==t_INT ? Fp_invsafe(z,p) : FpXQ_invsafe(z,T,p);
@@ -912,51 +1249,68 @@ FpXQXQ_div(GEN x,GEN y,GEN S, GEN T,GEN p)
   return gerepileupto(av, FpXQXQ_mul(x, FpXQXQ_inv(y,S,T,p),S,T,p));
 }
 
-typedef struct {
-  GEN T, S;
-  GEN p;
-} FpXQXQ_muldata;
-static GEN
-_FpXQXQ_add(void *data, GEN x, GEN y) {
-  FpXQXQ_muldata *d = (FpXQXQ_muldata*) data;
-  return FpXX_add(x,y, d->p);
-}
 static GEN
 _FpXQXQ_cmul(void *data, GEN P, long a, GEN x) {
-  FpXQXQ_muldata *d = (FpXQXQ_muldata*) data;
+  struct _FpXQXQ *d = (struct _FpXQXQ*) data;
   GEN y = gel(P,a+2);
   return typ(y)==t_INT ? FpXX_Fp_mul(x,y, d->p):
                          FpXX_FpX_mul(x,y,d->p);
 }
 static GEN
 _FpXQXQ_red(void *data, GEN x) {
-  FpXQXQ_muldata *d = (FpXQXQ_muldata*) data;
+  struct _FpXQXQ *d = (struct _FpXQXQ*) data;
   return FpXQX_red(x, d->T, d->p);
 }
 static GEN
 _FpXQXQ_mul(void *data, GEN x, GEN y) {
-  FpXQXQ_muldata *d = (FpXQXQ_muldata*) data;
+  struct _FpXQXQ *d = (struct _FpXQXQ*) data;
   return FpXQXQ_mul(x,y, d->S,d->T, d->p);
 }
 static GEN
 _FpXQXQ_sqr(void *data, GEN x) {
-  FpXQXQ_muldata *d = (FpXQXQ_muldata*) data;
+  struct _FpXQXQ *d = (struct _FpXQXQ*) data;
   return FpXQXQ_sqr(x, d->S,d->T, d->p);
 }
 
 static GEN
 _FpXQXQ_one(void *data) {
-  FpXQXQ_muldata *d = (FpXQXQ_muldata*) data;
-  return pol_1(varn(d->S));
+  struct _FpXQXQ *d = (struct _FpXQXQ*) data;
+  return pol_1(get_FpXQX_var(d->S));
 }
 
 static GEN
 _FpXQXQ_zero(void *data) {
-  FpXQXQ_muldata *d = (FpXQXQ_muldata*) data;
-  return pol_0(varn(d->S));
+  struct _FpXQXQ *d = (struct _FpXQXQ*) data;
+  return pol_0(get_FpXQX_var(d->S));
 }
 
-static struct bb_algebra FpXQXQ_algebra = { _FpXQXQ_red,_FpXQXQ_add,_FpXQXQ_mul,_FpXQXQ_sqr,_FpXQXQ_one,_FpXQXQ_zero };
+static struct bb_algebra FpXQXQ_algebra = { _FpXQXQ_red, _FpXQX_add,
+       _FpXQX_sub, _FpXQXQ_mul, _FpXQXQ_sqr, _FpXQXQ_one, _FpXQXQ_zero };
+
+const struct bb_algebra *
+get_FpXQXQ_algebra(void **E, GEN S, GEN T, GEN p)
+{
+  GEN z = new_chunk(sizeof(struct _FpXQXQ));
+  struct _FpXQXQ *e = (struct _FpXQXQ *) z;
+  e->T = FpX_get_red(T, p);
+  e->S = FpXQX_get_red(S, e->T, p);
+  e->p  = p; *E = (void*)e;
+  return &FpXQXQ_algebra;
+}
+
+static struct bb_algebra FpXQX_algebra = { _FpXQXQ_red, _FpXQX_add,
+       _FpXQX_sub, _FpXQX_mul, _FpXQX_sqr, _FpXQXQ_one, _FpXQXQ_zero };
+
+const struct bb_algebra *
+get_FpXQX_algebra(void **E, GEN T, GEN p, long v)
+{
+  GEN z = new_chunk(sizeof(struct _FpXQXQ));
+  struct _FpXQXQ *e = (struct _FpXQXQ *) z;
+  e->T = FpX_get_red(T, p);
+  e->S = pol_x(v);
+  e->p  = p; *E = (void*)e;
+  return &FpXQX_algebra;
+}
 
 /* x over Fq, return lift(x^n) mod S */
 GEN
@@ -964,7 +1318,7 @@ FpXQXQ_pow(GEN x, GEN n, GEN S, GEN T, GEN p)
 {
   pari_sp ltop = avma;
   GEN y;
-  FpXQXQ_muldata D;
+  struct _FpXQXQ D;
   long s = signe(n);
   if (!s) return pol_1(varn(x));
   if (is_pm1(n)) /* +/- 1 */
@@ -977,6 +1331,8 @@ FpXQXQ_pow(GEN x, GEN n, GEN S, GEN T, GEN p)
   }
   else
   {
+    T = FpX_get_red(T, p);
+    S = FpXQX_get_red(S, T, p);
     D.S = S; D.T = T; D.p = p;
     if (s < 0) x = FpXQXQ_inv(x,S,T,p);
     y = gen_pow(x, n, (void*)&D,&_FpXQXQ_sqr,&_FpXQXQ_mul);
@@ -988,8 +1344,10 @@ FpXQXQ_pow(GEN x, GEN n, GEN S, GEN T, GEN p)
 GEN
 FpXQXQ_powers(GEN x, long l, GEN S, GEN T, GEN p)
 {
-  FpXQXQ_muldata D;
-  int use_sqr = (degpol(x)<<1) >= degpol(S);
+  struct _FpXQXQ D;
+  int use_sqr = 2*degpol(x) >= get_FpXQX_degree(S);
+  T = FpX_get_red(T, p);
+  S = FpXQX_get_red(S, T, p);
   D.S = S; D.T = T; D.p = p;
   return gen_powers(x, l, use_sqr, (void*)&D, &_FpXQXQ_sqr, &_FpXQXQ_mul,&_FpXQXQ_one);
 }
@@ -1003,7 +1361,9 @@ FpXQXQ_matrix_pow(GEN y, long n, long m, GEN S, GEN T, GEN p)
 GEN
 FpXQX_FpXQXQV_eval(GEN P, GEN V, GEN S, GEN T, GEN p)
 {
-  FpXQXQ_muldata D;
+  struct _FpXQXQ D;
+  T = FpX_get_red(T, p);
+  S = FpXQX_get_red(S, T, p);
   D.S=S; D.T=T; D.p=p;
   return gen_bkeval_powers(P, degpol(P), V, (void*)&D, &FpXQXQ_algebra,
                                                    _FpXQXQ_cmul);
@@ -1012,58 +1372,105 @@ FpXQX_FpXQXQV_eval(GEN P, GEN V, GEN S, GEN T, GEN p)
 GEN
 FpXQX_FpXQXQ_eval(GEN Q, GEN x, GEN S, GEN T, GEN p)
 {
-  FpXQXQ_muldata D;
-  int use_sqr = (degpol(x)<<1) >= degpol(S);
+  struct _FpXQXQ D;
+  int use_sqr = 2*degpol(x) >= get_FpXQX_degree(S);
+  T = FpX_get_red(T, p);
+  S = FpXQX_get_red(S, T, p);
   D.S=S; D.T=T; D.p=p;
   return gen_bkeval(Q, degpol(Q), x, use_sqr, (void*)&D, &FpXQXQ_algebra,
-                                                    _FpXQXQ_cmul);
+      _FpXQXQ_cmul);
 }
 
 static GEN
-FpXQXQ_autpow_sqr(void * T, GEN x)
+FpXQXQ_autpow_sqr(void * E, GEN x)
 {
-  FpXQXQ_muldata *D = (FpXQXQ_muldata *)T;
+  struct _FpXQXQ *D = (struct _FpXQXQ *)E;
+  GEN T = D->T, p = D->p;
   GEN phi = gel(x,1), S = gel(x,2);
-  GEN phi2 = FpX_FpXQ_eval(phi,phi,D->T,D->p);
-  GEN Sphi = FpXY_FpXQ_evalx(S,phi,D->T,D->p);
-  GEN S2 = FpXQX_FpXQXQ_eval(Sphi, S, D->S,D->T,D->p);
+  long n = brent_kung_optpow(get_FpX_degree(T)-1,lgpol(S)+1,1);
+  GEN V = FpXQ_powers(phi, n, T, p);
+  GEN phi2 = FpX_FpXQV_eval(phi, V, T, p);
+  GEN Sphi = FpXY_FpXQV_evalx(S, V, T, p);
+  GEN S2 = FpXQX_FpXQXQ_eval(Sphi, S, D->S, T, p);
   return mkvec2(phi2, S2);
 }
 
 static GEN
-FpXQXQ_autpow_mul(void * T, GEN x, GEN y)
+FpXQXQ_autpow_mul(void * E, GEN x, GEN y)
 {
-  FpXQXQ_muldata *D = (FpXQXQ_muldata *)T;
+  struct _FpXQXQ *D = (struct _FpXQXQ *)E;
+  GEN T = D->T, p = D->p;
   GEN phi1 = gel(x,1), S1 = gel(x,2);
   GEN phi2 = gel(y,1), S2 = gel(y,2);
-  GEN phi3 = FpX_FpXQ_eval(phi1,phi2,D->T,D->p);
-  GEN Sphi = FpXY_FpXQ_evalx(S1,phi2,D->T,D->p);
-  GEN S3 = FpXQX_FpXQXQ_eval(Sphi, S2, D->S,D->T,D->p);
+  long n = brent_kung_optpow(get_FpX_degree(T)-1, lgpol(S1)+1, 1);
+  GEN V = FpXQ_powers(phi2, n, T, p);
+  GEN phi3 = FpX_FpXQV_eval(phi1, V, T, p);
+  GEN Sphi = FpXY_FpXQV_evalx(S1, V, T, p);
+  GEN S3 = FpXQX_FpXQXQ_eval(Sphi, S2, D->S, T, p);
   return mkvec2(phi3, S3);
 }
 
 GEN
 FpXQXQV_autpow(GEN aut, long n, GEN S, GEN T, GEN p)
 {
-  FpXQXQ_muldata D;
+  struct _FpXQXQ D;
+  T = FpX_get_red(T, p);
+  S = FpXQX_get_red(S, T, p);
   D.S=S; D.T=T; D.p=p;
   return gen_powu(aut,n,&D,FpXQXQ_autpow_sqr,FpXQXQ_autpow_mul);
 }
 
 static GEN
-FpXQXQ_autsum_mul(void * T, GEN x, GEN y)
+FpXQXQ_auttrace_mul(void *E, GEN x, GEN y)
 {
-  FpXQXQ_muldata *D = (FpXQXQ_muldata *)T;
+  struct _FpXQXQ *D = (struct _FpXQXQ *) E;
+  GEN T = D->T, p = D->p;
   GEN phi1 = gel(x,1), S1 = gel(x,2), a1 = gel(x,3);
   GEN phi2 = gel(y,1), S2 = gel(y,2), a2 = gel(y,3);
-  GEN phi3 = FpX_FpXQ_eval(phi1,phi2,D->T,D->p);
-  GEN Sphi = FpXY_FpXQ_evalx(S1,phi2,D->T,D->p);
-  long n = brent_kung_optpow(degpol(D->S)-1,2,1);
-  GEN V = FpXQXQ_powers(S2, n, D->S,D->T,D->p);
-  GEN S3 = FpXQX_FpXQXQV_eval(Sphi, V, D->S,D->T,D->p);
-  GEN aphi = FpXY_FpXQ_evalx(a1,phi2,D->T,D->p);
-  GEN aS = FpXQX_FpXQXQV_eval(aphi,V,D->S,D->T,D->p);
-  GEN a3 = FpXQXQ_mul(aS,a2,D->S,D->T,D->p);
+  long n2 = brent_kung_optpow(get_FpX_degree(T)-1, lgpol(S1)+lgpol(a1)+1, 1);
+  GEN V2 = FpXQ_powers(phi2, n2, T, p);
+  GEN phi3 = FpX_FpXQV_eval(phi1, V2, T, p);
+  GEN Sphi = FpXY_FpXQV_evalx(S1, V2, T, p);
+  GEN aphi = FpXY_FpXQV_evalx(a1, V2, T, p);
+  long n = brent_kung_optpow(maxss(degpol(Sphi),degpol(aphi)),2,1);
+  GEN V = FpXQXQ_powers(S2, n, D->S, T, p);
+  GEN S3 = FpXQX_FpXQXQV_eval(Sphi, V, D->S, T, p);
+  GEN aS = FpXQX_FpXQXQV_eval(aphi, V, D->S, T, p);
+  GEN a3 = FpXX_add(aS, a2, p);
+  return mkvec3(phi3, S3, a3);
+}
+
+static GEN
+FpXQXQ_auttrace_sqr(void *E, GEN x)
+{ return FpXQXQ_auttrace_mul(E, x, x); }
+
+GEN
+FpXQXQV_auttrace(GEN aut, long n, GEN S, GEN T, GEN p)
+{
+  struct _FpXQXQ D;
+  T = FpX_get_red(T, p);
+  S = FpXQX_get_red(S, T, p);
+  D.S=S; D.T=T; D.p=p;
+  return gen_powu(aut,n,&D,FpXQXQ_auttrace_sqr,FpXQXQ_auttrace_mul);
+}
+
+static GEN
+FpXQXQ_autsum_mul(void *E, GEN x, GEN y)
+{
+  struct _FpXQXQ *D = (struct _FpXQXQ *) E;
+  GEN T = D->T, p = D->p;
+  GEN phi1 = gel(x,1), S1 = gel(x,2), a1 = gel(x,3);
+  GEN phi2 = gel(y,1), S2 = gel(y,2), a2 = gel(y,3);
+  long n2 = brent_kung_optpow(get_FpX_degree(T)-1, lgpol(S1)+lgpol(a1)+1, 1);
+  GEN V2 = FpXQ_powers(phi2, n2, T, p);
+  GEN phi3 = FpX_FpXQV_eval(phi1, V2, T, p);
+  GEN Sphi = FpXY_FpXQV_evalx(S1, V2, T, p);
+  GEN aphi = FpXY_FpXQV_evalx(a1, V2, T, p);
+  long n = brent_kung_optpow(maxss(degpol(Sphi),degpol(aphi)),2,1);
+  GEN V = FpXQXQ_powers(S2, n, D->S, T, p);
+  GEN S3 = FpXQX_FpXQXQV_eval(Sphi, V, D->S, T, p);
+  GEN aS = FpXQX_FpXQXQV_eval(aphi, V, D->S, T, p);
+  GEN a3 = FpXQXQ_mul(aS,a2,D->S, T, p);
   return mkvec3(phi3, S3, a3);
 }
 
@@ -1074,7 +1481,9 @@ FpXQXQ_autsum_sqr(void * T, GEN x)
 GEN
 FpXQXQV_autsum(GEN aut, long n, GEN S, GEN T, GEN p)
 {
-  FpXQXQ_muldata D;
+  struct _FpXQXQ D;
+  T = FpX_get_red(T, p);
+  S = FpXQX_get_red(S, T, p);
   D.S=S; D.T=T; D.p=p;
   return gen_powu(aut,n,&D,FpXQXQ_autsum_sqr,FpXQXQ_autsum_mul);
 }
