@@ -96,7 +96,7 @@ hnffinal(GEN matgen,GEN perm,GEN* ptdep,GEN* ptB,GEN* ptC)
 {
   GEN p1,p2,U,H,Hnew,Bnew,Cnew,diagH1;
   GEN B = *ptB, C = *ptC, dep = *ptdep, depnew;
-  pari_sp av, lim;
+  pari_sp av;
   long i,j,k,s,i1,j1,zc;
   long co = lg(C);
   long col = lg(matgen)-1;
@@ -116,7 +116,7 @@ hnffinal(GEN matgen,GEN perm,GEN* ptdep,GEN* ptB,GEN* ptC)
 
   diagH1 = new_chunk(lnz+1); /* diagH1[i] = 0 iff H[i,i] != 1 (set later) */
 
-  av = avma; lim = stack_lim(av,1);
+  av = avma;
   Cnew = cgetg(co, typ(C));
   setlg(C, col+1); p1 = gmul(C,U);
   for (j=1; j<=col; j++) gel(Cnew,j) = gel(p1,j);
@@ -138,7 +138,7 @@ hnffinal(GEN matgen,GEN perm,GEN* ptdep,GEN* ptB,GEN* ptC)
       for (   ; k<=lig;  k++) gel(z,k) = subii(gel(z,k), mulii(p1, gel(Hi,k-nlze)));
       gel(Cnew,j) = gsub(gel(Cnew,j), gmul(p1, gel(Cnew,i+zc)));
     }
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,3))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"hnffinal, i = %ld",i);
       gerepileall(av, 2, &Cnew, &B);
@@ -217,6 +217,25 @@ col_dup(long l, GEN col)
   memcpy(c,col,l * sizeof(long)); return c;
 }
 
+/* permutation giving imagecompl(x') | image(x'), x' = transpose of x */
+static GEN
+ZM_rowrankprofile(GEN x, long *nlze)
+{
+  pari_sp av = avma;
+  GEN d, y;
+  long i, j, k, l, r;
+
+  x = shallowtrans(x); l = lg(x);
+  (void)new_chunk(l); /* HACK */
+  d = ZM_pivots(x,&r); avma = av;
+  *nlze = r;
+  if (!d) return identity_perm(l-1);
+  y = cgetg(l,t_VECSMALL);
+  for (i = j = 1, k = r+1; i<l; i++)
+    if (d[i]) y[k++] = i; else y[j++] = i;
+  return y;
+}
+
 /* HNF reduce a relation matrix (column operations + row permutation)
 ** Input:
 **   mat = (li-1) x (co-1) matrix of long
@@ -233,7 +252,7 @@ col_dup(long l, GEN col)
 GEN
 hnfspec_i(GEN mat0, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, long k0)
 {
-  pari_sp av, lim;
+  pari_sp av;
   long co, n, s, nlze, lnz, nr, i, j, k, lk0, col, lig, *p;
   GEN mat;
   GEN p1, p2, matb, matbnew, vmax, matt, T, extramat, B, C, H, dep, permpro;
@@ -262,7 +281,7 @@ hnfspec_i(GEN mat0, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, long k0)
     p1 = cgetg(k0+1,t_COL); gel(matt,j) = p1; gel(mat,j) = matj;
     for (i=1; i<=k0; i++) gel(p1,i) = stoi(matj[perm[i]]);
   }
-  av = avma; lim = stack_lim(av,1);
+  av = avma;
 
   i = lig = li-1; col = co-1; lk0 = k0;
   T = (k0 || (lg(C) > 1 && lgcols(C) > 1))? matid(col): NULL;
@@ -328,7 +347,7 @@ hnfspec_i(GEN mat0, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, long k0)
       if (T) ZC_lincomb1_inplace(gel(T,j), gel(T,col), stoi(-t));
     }
     lig--; col--;
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,3))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"hnfspec[1]");
       if (T) T = gerepilecopy(av, T); else avma = av;
@@ -371,7 +390,7 @@ hnfspec_i(GEN mat0, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, long k0)
       if (T) ZC_lincomb1_inplace(gel(T,j), gel(T,col), stoi(-t));
     }
     lig--; col--;
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,3))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"hnfspec[2]");
       gerepileall(av, T? 2: 1, &vmax, &T);
@@ -414,7 +433,7 @@ END2: /* clean up mat: remove everything to the right of the 1s on diagonal */
         for (h=1; h<i0; h++) gel(Bj,h) = subii(gel(Bj,h), mulii(v,gel(Bk,h)));
       }
       if (T) ZC_lincomb1_inplace(gel(T,j), gel(T,k), negi(v));
-      if (low_stack(lim, stack_lim(av,1)))
+      if (gc_needed(av,3))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"hnfspec[3], (i,j) = %ld,%ld", i,j);
         for (h=1; h<co; h++) setlg(matb[h], i0+1); /* bottom can be forgotten */
@@ -444,7 +463,7 @@ END2: /* clean up mat: remove everything to the right of the 1s on diagonal */
     nr = lnz;
   }
   else
-    permpro = ZM_imagecomplspec(extramat, &nr);
+    permpro = ZM_rowrankprofile(extramat, &nr);
   /* lnz = lg(permpro) */
   if (nlze)
   { /* put the nlze 0 rows (trivial generators) at the top */
@@ -625,10 +644,10 @@ hnfadd_i(GEN H, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, /* cf hnfspec */
   extramat = shallowconcat(extratop, vconcat(dep, H));
   Cnew     = shallowconcat(extraC, vecslice(C, col-lH+1, co));
   if (DEBUGLEVEL>5) err_printf("    1st phase done\n");
-  permpro = ZM_imagecomplspec(extramat, &nlze);
+  permpro = ZM_rowrankprofile(extramat, &nlze);
   extramat = rowpermute(extramat, permpro);
   *ptB     = rowpermute(B,        permpro);
-  permpro = vecpermute(perm, permpro);
+  permpro = vecsmallpermute(perm, permpro);
   for (i=1; i<=lig; i++) perm[i] = permpro[i]; /* perm o= permpro */
 
   *ptdep  = rowslice(extramat, 1, nlze);
@@ -715,7 +734,7 @@ gbezout_step(GEN *pa, GEN *pb, GEN *pu, GEN *pv, long vx)
   { /* possible accuracy problem */
     GEN D = RgX_gcd_simple(a,b);
     if (degpol(D)) {
-      D = RgX_Rg_div(D, leading_term(D));
+      D = RgX_Rg_div(D, leading_coeff(D));
       a = RgX_div(a, D);
       b = RgX_div(b, D);
       d = RgX_extgcd(a,b, pu,pv); /* retry now */
@@ -812,14 +831,14 @@ ZM_reduce(GEN A, GEN U, long i, long j0)
 }
 
 static GEN
-RgX_normalize(GEN T, GEN *pd)
+RgX_normalize_all(GEN T, GEN *pd)
 {
-  GEN d = leading_term(T);
+  GEN d = leading_coeff(T);
   while (gequal0(d) || ( typ(d) == t_REAL && lg(d) == 3
                        && gexpo(T) - expo(d) > (long)BITS_IN_LONG)) {
      T = normalizepol_lg(T, lg(T)-1);
      if (!signe(T)) { *pd = gen_1; return T; }
-     d = leading_term(T);
+     d = leading_coeff(T);
   }
   *pd = d;
   return RgX_Rg_div(T, d);
@@ -831,7 +850,7 @@ RgM_reduce(GEN A, GEN U, long i, long j0, long vx)
   long j, lA = lg(A);
   GEN d, T = gcoeff(A,i,j0);
   if (is_RgX(T,vx)) {
-    T = RgX_normalize(T, &d);
+    T = RgX_normalize_all(T, &d);
     if (degpol(T) == 0) { d = gel(T,2); T = gen_1; }
   } else {
     d = T; T = gen_1;
@@ -868,7 +887,7 @@ hnfmerge_get_1(GEN A, GEN B)
   b = gcoeff(B,1,1); lb = lgefint(b);
   if (!signe(b)) {
     if (!is_pm1(gcoeff(A,1,1))) return NULL;
-    return scalarcol_shallow(gen_1, l-1);
+    return gen_1;
   }
   for (j = 1; j < l; j++)
   {
@@ -922,12 +941,12 @@ remove_0cols(long r, GEN *pA, GEN *pB, long remove)
 static GEN
 hnf_i(GEN A, int remove)
 {
-  pari_sp av0 = avma, av, lim;
+  pari_sp av0 = avma, av;
   long s, n, m, j, k, li, def, ldef;
 
   RgM_dimensions(A, &m, &n);
   if (!n) return cgetg(1,t_MAT);
-  av = avma; lim = stack_lim(av,1);
+  av = avma;
   A = RgM_shallowcopy(A);
   def = n; ldef = (m>n)? m-n: 0;
   for (li=m; li>ldef; li--)
@@ -940,7 +959,7 @@ hnf_i(GEN A, int remove)
       /* zero a = Aij  using  b = Aik */
       k = (j==1)? def: j-1;
       ZC_elem(a,gcoeff(A,li,k), A,NULL, j,k);
-      if (low_stack(lim, stack_lim(av,1)))
+      if (gc_needed(av,1))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnf[1]. li=%ld",li);
         A = gerepilecopy(av, A);
@@ -955,7 +974,7 @@ hnf_i(GEN A, int remove)
     }
     else
       if (ldef) ldef--;
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnf[2]. li=%ld",li);
       A = gerepilecopy(av, A);
@@ -1000,12 +1019,12 @@ FpV_red_part_ipvec(GEN z, GEN p, long k)
 GEN
 ZpM_echelon(GEN x, long early_abort, GEN p, GEN pm)
 {
-  pari_sp av0 = avma, av, lim;
+  pari_sp av0 = avma, av;
   long m, li, co, i, j, k, def, ldef;
 
   co = lg(x); if (co == 1) return cgetg(1,t_MAT);
   li = lgcols(x);
-  av = avma; lim = stack_lim(av,1);
+  av = avma;
   x = RgM_shallowcopy(x);
   m = Z_pval(pm, p);
 
@@ -1048,7 +1067,7 @@ ZpM_echelon(GEN x, long early_abort, GEN p, GEN pm)
 
       t = diviiexact(a, pvmin); togglesign(t);
       ZC_lincomb1_inplace(gel(x,j), gel(x,def), t);
-      if (low_stack(lim, stack_lim(av,1)))
+      if (gc_needed(av,1))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"ZpM_echelon. i=%ld",i);
         x = gerepilecopy(av, x); pvmin = gcoeff(x,i,def);
@@ -1105,7 +1124,7 @@ zlm_echelon(GEN x, long early_abort, ulong p, ulong pm)
     /* pivot has valuation vmin */
     umin %= q;
     if (umin != 1)
-      Flc_Fl_mul_part_inplace(gel(x,def), Fl_inv(umin,q), pm, i-1);
+      Flv_Fl_mul_part_inplace(gel(x,def), Fl_inv(umin,q), pm, i-1);
     ucoeff(x, i, def) = pvmin = upowuu(p, vmin);
     for (j = def-1; j; j--)
     { /* zero x[i, 1..def-1] using x[i,def] = pvmin */
@@ -1133,7 +1152,7 @@ zlm_echelon(GEN x, long early_abort, ulong p, ulong pm)
 GEN
 ZM_hnfmodall_i(GEN x, GEN dm, long flag)
 {
-  pari_sp av, lim;
+  pari_sp av;
   const long center = (flag & hnf_CENTER);
   long moddiag = (flag & hnf_MODID);
   long li, co, i, j, k, def, ldef;
@@ -1161,7 +1180,7 @@ ZM_hnfmodall_i(GEN x, GEN dm, long flag)
       LDM[i] = lgefint(gel(dm,i));
     }
   }
-  av = avma; lim = stack_lim(av,1);
+  av = avma;
   x = RgM_shallowcopy(x);
 
   ldef = 0;
@@ -1194,11 +1213,11 @@ ZM_hnfmodall_i(GEN x, GEN dm, long flag)
         if (lgefint(gel(p2,k)) > LDM[k])
           gel(p2,k) = centermodii(gel(p2,k), gel(dm,k),gel(dm2,k));
       }
-      if (low_stack(lim, stack_lim(av,1)))
-      {
-        if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnfmod[1]. i=%ld",i);
-        x = gerepilecopy(av, x);
-      }
+    }
+    if (gc_needed(av,1))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnfmod[1]. i=%ld",i);
+      x = gerepilecopy(av, x);
     }
     if (moddiag && !signe(gcoeff(x,i,def)))
     { /* missing pivot on line i, insert column */
@@ -1236,11 +1255,11 @@ ZM_hnfmodall_i(GEN x, GEN dm, long flag)
         ZC_elem(a, gcoeff(x,j,j), x, NULL, li,j);
         FpV_red_part_ipvec(gel(x,li), dm, j-1);
         FpV_red_part_ipvec(gel(x,j),  dm, j-1);
-        if (low_stack(lim, stack_lim(av,1)))
-        {
-          if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnfmod[2]. i=%ld", i);
-          x = gerepilecopy(av, x);
-        }
+      }
+      if (gc_needed(av,1))
+      {
+        if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnfmod[2]. i=%ld", i);
+        x = gerepilecopy(av, x);
       }
     }
   }
@@ -1279,11 +1298,11 @@ ZM_hnfmodall_i(GEN x, GEN dm, long flag)
       p1 = gel(x,j);
       for (k=1; k<i; k++)
         if (lgefint(gel(p1,k)) > LDM[k]) gel(p1,k) = remii(gel(p1,k), gel(dm,i));
-      if (low_stack(lim, stack_lim(av,1)))
-      {
-        if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnfmod[3]. i=%ld", i);
-        gerepileall(av, 2, &x, &dm); diag = gcoeff(x,i,i);
-      }
+    }
+    if (gc_needed(av,1))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnfmod[3]. i=%ld", i);
+      gerepileall(av, 2, &x, &dm); diag = gcoeff(x,i,i);
     }
   }
   return x;
@@ -1331,7 +1350,7 @@ GEN
 ZM_hnfcenter(GEN M)
 {
   long i, j, k, N = lg(M)-1;
-  pari_sp av = avma, lim = stack_lim(av,1);
+  pari_sp av = avma;
 
   for (j=N-1; j>0; j--) /* skip last line */
   {
@@ -1350,9 +1369,9 @@ ZM_hnfcenter(GEN M)
       }
       else
         for (i = 1; i <= j; i++) gel(Mk,i) = subii(gel(Mk,i), mulii(q,gel(Mj,i)));
-      if (low_stack(lim, stack_lim(av,1)))
+      if (gc_needed(av,1))
       {
-        if (DEBUGMEM) pari_warn(warnmem,"ZM_hnfcenter, j = %ld",j);
+        if (DEBUGMEM>1) pari_warn(warnmem,"ZM_hnfcenter, j = %ld",j);
         M = gerepilecopy(av, M);
       }
     }
@@ -1407,7 +1426,7 @@ reduce2(GEN A, GEN B, long k, long j, long *row0, long *row1, GEN lambda, GEN D)
   *row1 = findi_normalize(gel(A,k), B,k,lambda);
   if (*row0)
     q = truedivii(gcoeff(A,*row0,k), gcoeff(A,*row0,j));
-  else if (absi_cmp(shifti(gcoeff(lambda,j,k), 1), gel(D,j)) > 0)
+  else if (abscmpii(shifti(gcoeff(lambda,j,k), 1), gel(D,j)) > 0)
     q = diviiround(gcoeff(lambda,j,k), gel(D,j));
   else
     return;
@@ -1485,7 +1504,7 @@ reverse_rows(GEN A)
 GEN
 ZM_hnflll(GEN A, GEN *ptB, int remove)
 {
-  pari_sp av = avma, lim = stack_lim(av,3);
+  pari_sp av = avma;
 #ifdef HNFLLL_QUALITY
   const long m1 = 1, n1 = 1; /* alpha = m1/n1. Maybe 3/4 here ? */
 #endif
@@ -1530,20 +1549,20 @@ ZM_hnflll(GEN A, GEN *ptB, int remove)
       {
         long row0, row1;
         reduce2(A,B,k,i,&row0,&row1,lambda,D);
-        if (low_stack(lim, stack_lim(av,3)))
+        if (gc_needed(av,3))
         {
           GEN b = D-1;
-          if (DEBUGMEM) pari_warn(warnmem,"hnflll (reducing), kmax = %ld",kmax);
+          if (DEBUGMEM>1) pari_warn(warnmem,"hnflll (reducing), kmax = %ld",kmax);
           gerepileall(av, B? 4: 3, &A, &lambda, &b, &B);
           D = b+1;
         }
       }
       if (++k > kmax) kmax = k;
     }
-    if (low_stack(lim, stack_lim(av,3)))
+    if (gc_needed(av,3))
     {
       GEN b = D-1;
-      if (DEBUGMEM) pari_warn(warnmem,"hnflll, kmax = %ld / %ld",kmax,n-1);
+      if (DEBUGMEM>1) pari_warn(warnmem,"hnflll, kmax = %ld / %ld",kmax,n-1);
       gerepileall(av, B? 4: 3, &A, &lambda, &b, &B);
       D = b+1;
     }
@@ -1581,7 +1600,7 @@ reduce1(GEN A, GEN B, long k, long j, GEN lambda, GEN D)
 
   if (signe(gel(A,j)))
     q = diviiround(gel(A,k),gel(A,j));
-  else if (absi_cmp(shifti(gcoeff(lambda,j,k), 1), gel(D,j)) > 0)
+  else if (abscmpii(shifti(gcoeff(lambda,j,k), 1), gel(D,j)) > 0)
     q = diviiround(gcoeff(lambda,j,k), gel(D,j));
   else
     return;
@@ -1652,12 +1671,12 @@ ZV_gcdext_i(GEN A)
   return mkvec2(gel(A,n-1), B);
 }
 GEN
-ZV_gcdext(GEN A)
+ZV_extgcd(GEN A)
 {
   pari_sp av = avma;
   return gerepilecopy(av, ZV_gcdext_i(A));
 }
-/* as ZV_gcdext, transforming the gcd into a t_MAT, for mathnf0 */
+/* as ZV_extgcd, transforming the gcd into a t_MAT, for mathnf0 */
 static GEN
 ZV_hnfgcdext(GEN A)
 {
@@ -1674,7 +1693,7 @@ GEN
 ZM_hnfperm(GEN A, GEN *ptU, GEN *ptperm)
 {
   GEN U, c, l, perm, d, p, q, b;
-  pari_sp av = avma, av1, lim;
+  pari_sp av = avma, av1;
   long r, t, i, j, j1, k, m, n;
 
   n = lg(A)-1;
@@ -1688,7 +1707,7 @@ ZM_hnfperm(GEN A, GEN *ptU, GEN *ptperm)
   c = zero_zv(m);
   l = zero_zv(n);
   perm = cgetg(m+1, t_VECSMALL);
-  av1 = avma; lim = stack_lim(av1,1);
+  av1 = avma;
   A = RgM_shallowcopy(A);
   U = ptU? matid(n): NULL;
   /* U base change matrix : A0*U = A all along */
@@ -1726,7 +1745,7 @@ ZM_hnfperm(GEN A, GEN *ptU, GEN *ptperm)
       for (i=t-1; i; i--)
       {
         q = gcoeff(A,i,k);
-        if (signe(q) && absi_cmp(p,q) > 0) { p = q; t = i; }
+        if (signe(q) && abscmpii(p,q) > 0) { p = q; t = i; }
       }
       perm[++r] = l[k] = t; c[t] = k;
       if (signe(p) < 0)
@@ -1747,9 +1766,9 @@ ZM_hnfperm(GEN A, GEN *ptU, GEN *ptperm)
         if (U) ZC_lincomb1_inplace(gel(U,j), gel(U,k), q);
       }
     }
-    if (low_stack(lim, stack_lim(av1,1)))
+    if (gc_needed(av1,1))
     {
-      if (DEBUGMEM>1) pari_warn(warnmem,"hnfperm");
+      if (DEBUGMEM>1) pari_warn(warnmem,"hnfperm, k=%ld",k);
       gerepileall(av1, U? 2: 1, &A, &U);
     }
   }
@@ -1789,6 +1808,27 @@ ZM_hnfperm(GEN A, GEN *ptU, GEN *ptperm)
   return p;
 }
 
+GEN
+ZM_hnf_knapsack(GEN x)
+{
+  GEN t, perm, H = ZM_hnfperm(x,NULL,&perm);
+  long i,j, l = lg(H), h = lgcols(H);
+  for (i=1; i<h; i++)
+  {
+    int fl = 0;
+    for (j=1; j<l; j++)
+    {
+      t = gcoeff(H,i,j);
+      if (signe(t))
+      {
+        if (!is_pm1(t) || fl) return NULL;
+        fl = 1;
+      }
+    }
+  }
+  return rowpermute(H, perm_inv(perm));
+}
+
 GEN
 hnfperm(GEN A)
 {
@@ -1801,9 +1841,9 @@ hnfperm(GEN A)
  * If 'remove' = 1, remove 0 columns (do NOT update *ptB accordingly)
  * If 'remove' = 2, remove 0 columns and update *ptB accordingly */
 GEN
-ZM_hnfall(GEN A, GEN *ptB, long remove)
+ZM_hnfall_i(GEN A, GEN *ptB, long remove)
 {
-  pari_sp av = avma, av1, lim;
+  pari_sp av;
   long m, n, r, i, j, k, li;
   GEN B, c, h, a;
 
@@ -1815,7 +1855,7 @@ ZM_hnfall(GEN A, GEN *ptB, long remove)
   }
   c = zero_zv(m);
   h = const_vecsmall(n, m);
-  av1 = avma; lim = stack_lim(av1,1);
+  av = avma;
   A = RgM_shallowcopy(A);
   B = ptB? matid(n): NULL;
   r = n+1;
@@ -1830,10 +1870,10 @@ ZM_hnfall(GEN A, GEN *ptB, long remove)
         /* zero a = Aij  using  Aik */
         if (signe(a)) ZC_elem(a,gcoeff(A,i,k), A,B,j,k);
         ZM_reduce(A,B, i,k); /* ensure reduced entries */
-        if (low_stack(lim, stack_lim(av1,1)))
+        if (gc_needed(av,1))
         {
           if (DEBUGMEM>1) pari_warn(warnmem,"hnfall[1], li = %ld", li);
-          gerepileall(av1, B? 2: 1, &A, &B);
+          gerepileall(av, B? 2: 1, &A, &B);
         }
       }
       if (signe( gcoeff(A,li,j) )) break;
@@ -1853,10 +1893,10 @@ ZM_hnfall(GEN A, GEN *ptB, long remove)
       if (B) ZV_togglesign(gel(B,r));
     }
     ZM_reduce(A,B, li,r);
-    if (low_stack(lim, stack_lim(av1,1)))
+    if (gc_needed(av,1))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"hnfall[2], li = %ld", li);
-      gerepileall(av1, B? 2: 1, &A, &B);
+      gerepileall(av, B? 2: 1, &A, &B);
     }
   }
 
@@ -1869,16 +1909,23 @@ ZM_hnfall(GEN A, GEN *ptB, long remove)
       k = c[i];
       if (signe(a)) ZC_elem(a,gcoeff(A,i,k), A,B, j,k);
       ZM_reduce(A,B, i,k); /* ensure reduced entries, even if a = 0 */
-      if (low_stack(lim, stack_lim(av1,1)))
+      if (gc_needed(av,1))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"hnfall[3], j = %ld", j);
-        gerepileall(av1, B? 2: 1, &A, &B);
+        gerepileall(av, B? 2: 1, &A, &B);
       }
     }
   if (DEBUGLEVEL>5) err_printf("\n");
   if (remove) remove_0cols(r, &A, &B, remove);
-  gerepileall(av, B? 2: 1, &A, &B);
-  if (B) *ptB = B;
+  if (ptB) *ptB = B;
+  return A;
+}
+GEN
+ZM_hnfall(GEN A, GEN *ptB, long remove)
+{
+  pari_sp av = avma;
+  A = ZM_hnfall_i(A, ptB, remove);
+  gerepileall(av, ptB? 2: 1, &A, ptB);
   return A;
 }
 
@@ -1892,6 +1939,30 @@ hnfall(GEN x)
 GEN
 hnf(GEN x) { return mathnf0(x,0); }
 
+/* C = A^(-1)t where A and C are integral, A is upper triangular, t t_INT */
+GEN
+hnf_invscale(GEN A, GEN t)
+{
+  long n = lg(A)-1, i,j,k;
+  GEN m, c = cgetg(n+1,t_MAT);
+
+  if (!n) return c;
+  for (k=1; k<=n; k++)
+  { /* cf hnf_divscale with B = id, thus b = e_k */
+    GEN u = cgetg(n+1, t_COL);
+    pari_sp av = avma;
+    gel(c,k) = u;
+    gel(u,n) = k == n? gerepileuptoint(av, diviiexact(t, gcoeff(A,n,n))): gen_0;
+    for (i=n-1; i>0; i--)
+    {
+      av = avma; m = i == k? t: gen_0;
+      for (j=i+1; j<=n; j++) m = subii(m, mulii(gcoeff(A,i,j),gel(u,j)));
+      gel(u,i) = gerepileuptoint(av, diviiexact(m, gcoeff(A,i,i)));
+    }
+  }
+  return c;
+}
+
 /* C = A^(-1)(tB) where A, B, C are integral, A is upper triangular, t t_INT */
 GEN
 hnf_divscale(GEN A, GEN B, GEN t)
@@ -2038,7 +2109,7 @@ static GEN
 bezout_step(GEN *pa, GEN *pb, GEN *pu, GEN *pv)
 {
   GEN a = *pa, b = *pb, d;
-  if (absi_equal(a,b))
+  if (absequalii(a,b))
   {
     long sa = signe(a), sb = signe(b);
     *pv = gen_0;
@@ -2086,7 +2157,7 @@ ZM_snf_no_divide(GEN x, long i)
 GEN
 ZM_snfall_i(GEN x, GEN *ptU, GEN *ptV, int return_vec)
 {
-  pari_sp av0 = avma, av, lim = stack_lim(av0,1);
+  pari_sp av0 = avma, av;
   long i, j, k, m0, m, n0, n;
   GEN p1, u, v, U, V, V0, mdet, ys, perm = NULL;
 
@@ -2187,7 +2258,7 @@ ZM_snfall_i(GEN x, GEN *ptU, GEN *ptV, int return_vec)
         b = gcoeff(x,i,j); if (!signe(b)) continue;
         a = gcoeff(x,i,i);
         ZC_elem(b, a, x,V, j,i);
-        if (low_stack(lim, stack_lim(av,1)))
+        if (gc_needed(av,1))
         {
           if (DEBUGMEM>1) pari_warn(warnmem,"[1]: ZM_snfall i = %ld", i);
           snf_pile(av, &x,&U,&V);
@@ -2210,7 +2281,7 @@ ZM_snfall_i(GEN x, GEN *ptU, GEN *ptV, int return_vec)
         gcoeff(x,j,i) = gen_0;
         gcoeff(x,i,i) = d;
         if (U) update(u,v,a,b,(GEN*)(U+i),(GEN*)(U+j));
-        if (low_stack(lim, stack_lim(av,1)))
+        if (gc_needed(av,1))
         {
           if (DEBUGMEM>1) pari_warn(warnmem,"[2]: ZM_snfall, i = %ld", i);
           snf_pile(av, &x,&U,&V);
@@ -2227,7 +2298,7 @@ ZM_snfall_i(GEN x, GEN *ptU, GEN *ptV, int return_vec)
           gcoeff(x,i,j) = addii(gcoeff(x,i,j),gcoeff(x,k,j));
         if (U) gel(U,i) = gadd(gel(U,i),gel(U,k));
       }
-      if (low_stack(lim, stack_lim(av,1)))
+      if (gc_needed(av,1))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"[3]: ZM_snfall");
         snf_pile(av, &x,&U,&V);
@@ -2352,7 +2423,7 @@ smithclean(GEN z)
   {
     if (h > l)
     {
-      for (i=1+h-l, j=1; i<h; i++,j++)
+      for (i=1+h-l, j=1; i<c; i++,j++)
         gcoeff(t,i,j) = gcopy(gcoeff(D,i,j));
     }
     else if (h < l)
@@ -2405,7 +2476,7 @@ gsnf_no_divide(GEN x, long i, long vx)
 GEN
 RgM_hnfall(GEN A, GEN *pB, long remove)
 {
-  pari_sp av, lim;
+  pari_sp av;
   long li, j, k, m, n, def, ldef;
   GEN B;
   long vx = gvar(A);
@@ -2417,7 +2488,7 @@ RgM_hnfall(GEN A, GEN *pB, long remove)
     return ZM_hnfall(A, pB, remove);
   }
   m = nbrows(A);
-  av = avma; lim = stack_lim(av,1);
+  av = avma;
   A = RgM_shallowcopy(A);
   B = pB? matid(n): NULL;
   def = n; ldef = (m>n)? m-n: 0;
@@ -2438,12 +2509,12 @@ RgM_hnfall(GEN A, GEN *pB, long remove)
     else
     {
       GEN d;
-      gcoeff(A,li,def) = RgX_normalize(T, &d);
+      gcoeff(A,li,def) = RgX_normalize_all(T, &d);
       if (B && !gequal1(d)) gel(B, def) = RgC_Rg_div(gel(B, def), d);
       RgM_reduce(A, B, li, def, vx);
       def--;
     }
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"ghnfall");
       gerepileall(av, B? 2: 1, &A, &B);
@@ -2459,7 +2530,7 @@ RgM_hnfall(GEN A, GEN *pB, long remove)
 static GEN
 gsmithall_i(GEN x,long all)
 {
-  pari_sp av, lim;
+  pari_sp av;
   long i, j, k, n;
   GEN z, u, v, U, V;
   long vx = gvar(x);
@@ -2468,7 +2539,7 @@ gsmithall_i(GEN x,long all)
   n = lg(x)-1;
   if (!n) return trivsmith(all);
   if (lgcols(x) != n+1) pari_err_DIM("gsmithall");
-  av = avma; lim = stack_lim(av,1);
+  av = avma;
   x = RgM_shallowcopy(x);
   if (all) { U = matid(n); V = matid(n); }
   for (i=n; i>=2; i--)
@@ -2517,7 +2588,7 @@ gsmithall_i(GEN x,long all)
           gcoeff(x,i,j) = gadd(gcoeff(x,i,j),gcoeff(x,k,j));
         if (all) gel(U,i) = gadd(gel(U,i),gel(U,k));
       }
-      if (low_stack(lim, stack_lim(av,1)))
+      if (gc_needed(av,1))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"gsmithall");
         gerepileall(av, all? 3: 1, &x, &U, &V);
@@ -2528,7 +2599,7 @@ gsmithall_i(GEN x,long all)
   {
     GEN d, T = gcoeff(x,k,k);
     if (!signe(T)) continue;
-    if (is_RgX(T,vx)) T = RgX_normalize(T, &d); else { d = T; T = gen_1; }
+    if (is_RgX(T,vx)) T = RgX_normalize_all(T, &d); else { d = T; T = gen_1; }
     if (all && !gequal1(d)) gel(V,k) = RgC_Rg_div(gel(V,k), d);
     gcoeff(x,k,k) = T;
   }
@@ -2554,15 +2625,10 @@ gsmith(GEN x) { return gsmithall_i(x,0); }
 GEN
 gsmithall(GEN x) { return gsmithall_i(x,1); }
 
-/* H relation matrix among row of generators g in HNF.  Let URV = D its SNF,
- * newU R newV = newD its clean SNF (no 1 in Dnew). Return the diagonal of
- * newD, newU and newUi such that  1/U = (newUi, ?).
- * Rationale: let (G,0) = g Ui be the new generators then
- * 0 = G U R --> G D = 0,  g = G newU  and  G = g newUi */
-GEN
-ZM_snf_group(GEN H, GEN *newU, GEN *newUi)
+/* H is a relation matrix, either in HNF or a t_VEC (diagonal HNF) */
+static GEN
+snf_group(GEN H, GEN D, GEN *newU, GEN *newUi)
 {
-  GEN D = ZM_snfall_i(H, newU, newUi, 1);
   long i, j, l;
 
   ZM_snfclean(D, newU? *newU: NULL, newUi? *newUi: NULL);
@@ -2577,122 +2643,110 @@ ZM_snf_group(GEN H, GEN *newU, GEN *newUi)
     }
     *newU = U;
   }
-  if (newUi) { /* UHV=D -> U^-1 = HVD^-1 -> U^-1 = H(VD^-1 mod 1) mod H */
+  if (newUi) { /* UHV=D -> U^-1 = (HV)D^-1 -> U^-1 = H(VD^-1 mod 1) mod H */
     if (l > 1)
     { /* Ui = ZM_inv(U, gen_1); setlg(Ui, l); */
-      GEN V = FpM_red(*newUi, gel(D,1));
-      GEN Ui = ZM_mul(H, V);
+      GEN V = *newUi, Ui;
+      for (i = 1; i < l; i++) gel(V,i) = FpC_red(gel(V,i), gel(D,i));
+      Ui = typ(H) == t_VEC? ZM_diag_mul(H, V): ZM_mul(H, V);
       for (i = 1; i < l; i++) gel(Ui,i) = ZC_Z_divexact(gel(Ui,i), gel(D,i));
-      *newUi = ZM_hnfrem(Ui, H);
+      if (typ(H) == t_VEC)
+      { for (i = 1; i < l; i++) gel(Ui,i) = vecmodii(gel(Ui,i), H); }
+      else
+        Ui = ZM_hnfrem(Ui, H);
+      *newUi = Ui;
     }
   }
   return D;
 }
-
-/***********************************************************************
- ****                                                               ****
- ****         Frobenius form and Jordan form of a matrix            ****
- ****                                                               ****
- ***********************************************************************/
-
-static long
-prod_degree(GEN V)
-{
-  long i, s=0, l = lg(V);
-  for (i=1; i<l; i++)
-  {
-    long d = degpol(gel(V,i));
-    if (d<0) return d;
-    s += d;
-  }
-  return s;
-}
-
+/* H relation matrix among row of generators g in HNF.  Let URV = D its SNF,
+ * newU R newV = newD its clean SNF (no 1 in Dnew). Return the diagonal of
+ * newD, newU and newUi such that  1/U = (newUi, ?).
+ * Rationale: let (G,0) = g Ui be the new generators then
+ * 0 = G U R --> G D = 0,  g = G newU  and  G = g newUi */
 GEN
-Frobeniusform(GEN V, long n)
+ZM_snf_group(GEN H, GEN *newU, GEN *newUi)
 {
-  long i, j, k;
-  GEN M = zeromatcopy(n,n);
-  for (k=1,i=1;i<lg(V);i++,k++)
-  {
-    GEN  P = gel(V,i);
-    long d = degpol(P);
-    if (k+d-1 > n) pari_err_PREC("matfrobenius");
-    for (j=0; j<d-1; j++, k++) gcoeff(M,k+1,k) = gen_1;
-    for (j=0; j<d; j++) gcoeff(M,k-j,k) = gneg(gel(P, 1+d-j));
-  }
-  return M;
+  GEN D = ZM_snfall_i(H, newU, newUi, 1);
+  return snf_group(H, D, newU, newUi);
 }
 
-static GEN
-build_frobeniusbc(GEN V, long n)
+/* D a ZV: SNF for matdiagonal(D). Faster because we only ensure elementary
+ * divisors condition: d[n] | ... | d[1] and need not convert D to matrix form*/
+GEN
+ZV_snfall(GEN D, GEN *pU, GEN *pV)
 {
-  long i, j, k, l, m = lg(V)-1;
-  GEN M = zeromatcopy(n,n), z = monomial(gen_m1, 1, 0); /* -x */
-  for (k=1,l=1+m,i=1;i<=m;i++,k++)
+  pari_sp av = avma;
+  long j, n = lg(D)-1;
+  GEN U = pU? matid(n): NULL;
+  GEN V = pV? matid(n): NULL;
+  GEN p;
+
+  D = leafcopy(D);
+  for (j = n; j > 0; j--)
   {
-    GEN  P = gel(V,i);
-    long d = degpol(P);
-    if (d <= 0) continue;
-    if (l+d-2 > n) pari_err_PREC("matfrobenius");
-    gcoeff(M,k,i) = gen_1;
-    for (j=1; j<d; j++,k++,l++)
+    GEN b = gel(D,j);
+    if (signe(b) < 0)
     {
-      gcoeff(M,k,l)   = z;
-      gcoeff(M,k+1,l) = gen_1;
+      gel(D,j) = absi(b);
+      if (V) ZV_togglesign(gel(V,j));
     }
   }
-  return M;
-}
-
-static GEN
-build_basischange(GEN N, GEN U)
-{
-  long i, j, n = lg(N)-1;
-  GEN p2 = cgetg(n+1, t_MAT);
-  for (j = 1; j <= n; ++j)
+  /* entries are non-negative integers */
+  p = gen_indexsort(D, NULL, &negcmpii);
+  D = vecpermute(D, p);
+  if (U) U = vecpermute(U, p);
+  if (V) V = vecpermute(V, p);
+  /* entries are sorted by decreasing value */
+  for (j = n; j > 0; j--)
   {
-    pari_sp btop = avma;
-    GEN p3 = NULL;
-    for (i = 1; i <= n; ++i)
-    {
-      GEN Uij = gcoeff(U, i, j), z;
-      if (is_RgX(Uij, 0))
-        z = RgX_RgM_eval_col(Uij, N, i);
-      else
-        z = Rg_col_ei(Uij, n, i);
-      p3 = p3 ? RgC_add(p3, z):z;
+    GEN b = gel(D,j);
+    long i;
+    for (i = j-1; i > 0; i--)
+    { /* invariant: a >= b. If au+bv = d is a Bezout relation, A=a/d and B=b/d
+       * we have [B,-A;u,v]*diag(a,b)*[1-u*A,1; -u*A,1]] = diag(Ab, d) */
+      GEN a = gel(D,i), u,v, d = bezout(a,b, &u,&v), A, Wi, Wj;
+      if (equalii(d,b)) continue;
+      A = diviiexact(a,d);
+      if (V)
+      {
+        GEN t = mulii(u,A);
+        Wi = ZC_lincomb(subui(1,t), negi(t), gel(V,i), gel(V,j));
+        Wj = ZC_add(gel(V,i), gel(V,j));
+        gel(V,i) = Wi;
+        gel(V,j) = Wj;
+      }
+      if (U)
+      {
+        GEN B = diviiexact(b,d);
+        Wi = ZC_lincomb(B, negi(A), gel(U,i), gel(U,j));
+        Wj = ZC_lincomb(u, v, gel(U,i), gel(U,j));
+        gel(U,i) = Wi;
+        gel(U,j) = Wj;
+      }
+      gel(D,i) = mulii(A,b); /* lcm(a,b) */
+      gel(D,j) = d; /* gcd(a,b) */
+      b = gel(D,j); if (equali1(b)) break;
     }
-    gel(p2,j) = gerepileupto(btop, p3);
   }
-  return p2;
+  snf_pile(av, &D,&U,&V);
+  if (U) *pU = shallowtrans(U);
+  if (V) *pV = V;
+  return D;
+}
+GEN
+ZV_snf_group(GEN d, GEN *newU, GEN *newUi)
+{
+  GEN D = ZV_snfall(d, newU, newUi);
+  return snf_group(d, D, newU, newUi);
 }
 
-GEN
-matfrobenius(GEN M, long flag, long v)
+/* D a vector of elementary divisors. Truncate (setlg) to leave out trivial
+ * entries (= 1) */
+void
+ZV_snf_trunc(GEN D)
 {
-  pari_sp ltop=avma;
-  long n;
-  GEN D, A, N, B, R, M_x;
-  if (typ(M)!=t_MAT) pari_err_TYPE("matfrobenius",M);
-  if (v<0) v=0;
-  if (varncmp(gvar(M), v) <= 0) pari_err_PRIORITY("matfrobenius", M, "<=", v);
-  n = lg(M)-1;
-  if (n && lgcols(M)!=n+1) pari_err_DIM("matfrobenius");
-  M_x = RgM_Rg_add_shallow(M, monomial(gen_m1, 1, v));
-  if (flag<2)
-  {
-    D = matsnf0(M_x,6);
-    if (prod_degree(D) != n) pari_err_PREC("matfrobenius");
-    if (flag != 1) D = Frobeniusform(D, n);
-    return gerepileupto(ltop, D);
-  }
-  if (flag>2) pari_err_FLAG("matfrobenius");
-  A = matsnf0(M_x,3);
-  D = smithclean(RgM_diagonal_shallow(gel(A,3)));
-  if (prod_degree(D) != n) pari_err_PREC("matfrobenius");
-  N = Frobeniusform(D, n);
-  B = build_frobeniusbc(D, n);
-  R = build_basischange(N, RgM_mul(B,gel(A,1)));
-  return gerepilecopy(ltop, mkvec2(N,R));
+  long i, l = lg(D);
+  for (i = 1; i < l; i++)
+    if (is_pm1(gel(D,i))) { setlg(D,i); break; }
 }
