@@ -20,6 +20,10 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
   do { GEN _v = cgetg(3, t_FRAC);\
        gel(_v,1) = (x);\
        gel(_v,2) = (y); return _v; } while(0)
+#define retmkrfrac(x,y)\
+  do { GEN _v = cgetg(3, t_RFRAC);\
+       gel(_v,1) = (x);\
+       gel(_v,2) = (y); return _v; } while(0)
 #define retmkintmod(x,y)\
   do { GEN _v = cgetg(3, t_INTMOD);\
        gel(_v,1) = (y);\
@@ -98,6 +102,14 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
        gel(_v,3) = (z);\
        gel(_v,4) = (t);\
        gel(_v,5) = (u); return _v; } while(0)
+#define retmkcol6(x,y,z,t,u,v)\
+  do { GEN _v = cgetg(7, t_COL);\
+       gel(_v,1) = (x);\
+       gel(_v,2) = (y);\
+       gel(_v,3) = (z);\
+       gel(_v,4) = (t);\
+       gel(_v,5) = (u);\
+       gel(_v,6) = (v); return _v; } while(0)
 #define retmkmat(x)\
   do { GEN _v = cgetg(2, t_MAT);\
        gel(_v,1) = (x); return _v; } while(0)
@@ -165,6 +177,9 @@ mkvecsmall3(long x,long y,long z) { GEN v = cgetg(4, t_VECSMALL);
 INLINE GEN
 mkvecsmall4(long x,long y,long z,long t) { GEN v = cgetg(5, t_VECSMALL);
   v[1]=x; v[2]=y; v[3]=z; v[4]=t; return v; }
+INLINE GEN
+mkvecsmall5(long x,long y,long z,long t,long u) { GEN v = cgetg(6, t_VECSMALL);
+  v[1]=x; v[2]=y; v[3]=z; v[4]=t; v[5]=u; return v; }
 
 INLINE GEN
 mkqfi(GEN x, GEN y, GEN z) { retmkqfi(x,y,z); }
@@ -204,6 +219,8 @@ mkcol4(GEN x, GEN y, GEN z, GEN t) { retmkcol4(x,y,z,t); }
 INLINE GEN
 mkcol5(GEN x, GEN y, GEN z, GEN t, GEN u) { retmkcol5(x,y,z,t,u); }
 INLINE GEN
+mkcol6(GEN x, GEN y, GEN z, GEN t, GEN u, GEN v) { retmkcol6(x,y,z,t,u,v); }
+INLINE GEN
 mkcols(long x) { retmkcol(stoi(x)); }
 INLINE GEN
 mkcol2s(long x, long y) { retmkcol2(stoi(x),stoi(y)); }
@@ -228,6 +245,12 @@ INLINE GEN
 mkmatcopy(GEN x) { GEN v = cgetg(2, t_MAT); gel(v,1) = gcopy(x); return v; }
 INLINE GEN
 mkerr(long x) { GEN v = cgetg(2, t_ERROR); v[1] = x; return v; }
+INLINE GEN
+mkoo(void) { GEN v = cgetg(2, t_INFINITY); gel(v,1) = gen_1; return v; }
+INLINE GEN
+mkmoo(void) { GEN v = cgetg(2, t_INFINITY); gel(v,1) = gen_m1; return v; }
+INLINE long
+inf_get_sign(GEN x) { return signe(gel(x,1)); }
 /* pol */
 INLINE GEN
 pol_x(long v) {
@@ -236,6 +259,22 @@ pol_x(long v) {
   gel(p,2) = gen_0;
   gel(p,3) = gen_1; return p;
 }
+/* x^n, assume n >= 0 */
+INLINE GEN
+pol_xn(long n, long v) {
+  long i, a = n+2;
+  GEN p = cgetg(a+1, t_POL);
+  p[1] = evalsigne(1)|evalvarn(v);
+  for (i = 2; i < a; i++) gel(p,i) = gen_0;
+  gel(p,a) = gen_1; return p;
+}
+/* x^n, no assumption on n */
+INLINE GEN
+pol_xnall(long n, long v)
+{
+  if (n < 0) retmkrfrac(gen_1, pol_xn(-n,v));
+  return pol_xn(n, v);
+}
 INLINE GEN
 pol_1(long v) {
   GEN p = cgetg(3, t_POL);
@@ -300,6 +339,16 @@ zeroser(long v, long e)
   GEN x = cgetg(2, t_SER);
   x[1] = evalvalp(e) | evalvarn(v); return x;
 }
+INLINE int
+ser_isexactzero(GEN x)
+{
+  if (!signe(x)) switch(lg(x))
+  {
+    case 2: return 1;
+    case 3: return isexactzero(gel(x,2));
+  }
+  return 0;
+}
 /* 0 * pol_x(v) */
 INLINE GEN
 zeropol(long v) { return pol_0(v); }
@@ -466,6 +515,15 @@ vec_lengthen(GEN v, long n)
 }
 /* shallow*/
 INLINE GEN
+vec_append(GEN V, GEN s)
+{
+  long i, l2 = lg(V);
+  GEN res = cgetg(l2+1, typ(V));
+  for (i = 1; i < l2; ++i) gel(res, i) = gel(V,i);
+  gel(res,l2) = s; return res;
+}
+/* shallow*/
+INLINE GEN
 vec_setconst(GEN v, GEN x)
 {
   long i, l = lg(v);
@@ -546,12 +604,11 @@ vecsmall_prepend(GEN V, long s)
   return res;
 }
 
-/*Can be used on t_VEC, but coeffs not gcopy-ed*/
 INLINE GEN
 vecsmall_append(GEN V, long s)
 {
   long i, l2 = lg(V);
-  GEN res = cgetg(l2+1, typ(V));
+  GEN res = cgetg(l2+1, t_VECSMALL);
   for (i = 1; i < l2; ++i) res[i] = V[i];
   res[l2] = s; return res;
 }
@@ -737,14 +794,16 @@ INLINE void
 pari_stack_alloc(pari_stack *s, long nb)
 {
   void **sdat = pari_stack_base(s);
-  if (s->n+nb <= s->alloc) return;
-  if (!s->alloc)
-    s->alloc = nb;
+  long alloc = s->alloc;
+  if (s->n+nb <= alloc) return;
+  if (!alloc)
+    alloc = nb;
   else
   {
-    while (s->n+nb > s->alloc) s->alloc <<= 1;
+    while (s->n+nb > alloc) alloc <<= 1;
   }
-  *sdat = pari_realloc(*sdat,s->alloc*s->size);
+  *sdat = pari_realloc(*sdat,alloc*s->size);
+  s->alloc = alloc;
 }
 
 INLINE long
@@ -754,7 +813,7 @@ INLINE void
 pari_stack_delete(pari_stack *s)
 {
   void **sdat = pari_stack_base(s);
-  if (*sdat) free(*sdat);
+  if (*sdat) pari_free(*sdat);
 }
 
 INLINE void
@@ -804,6 +863,21 @@ rowslice(GEN A, long x1, long x2)
   return B;
 }
 
+INLINE GEN
+matslice(GEN A, long x1, long x2, long y1, long y2)
+{
+  return rowslice(vecslice(A, y1, y2), x1, x2);
+}
+
+/* shallow, remove coeff of index j */
+INLINE GEN
+rowsplice(GEN a, long j)
+{
+  long i, l;
+  GEN b = cgetg_copy(a,&l);
+  for (i = 1; i < l; i++) gel(b,i) = vecsplice(gel(a,i), j);
+  return b;
+}
 /* shallow, remove coeff of index j */
 INLINE GEN
 vecsplice(GEN a, long j)
@@ -836,7 +910,7 @@ row(GEN A, long x0)
   return B;
 }
 INLINE GEN
-row_Flm(GEN A, long x0)
+Flm_row(GEN A, long x0)
 {
   long i, lB = lg(A);
   GEN B  = cgetg(lB, t_VECSMALL);
@@ -865,12 +939,36 @@ row_i(GEN A, long x0, long x1, long x2)
 INLINE GEN
 vecreverse(GEN A)
 {
-  long i,l = lg(A);
-  GEN B = cgetg(l, typ(A));
+  long i, l;
+  GEN B = cgetg_copy(A, &l);
   for (i=1; i<l; i++) gel(B, i) = gel(A, l-i);
   return B;
 }
 
+INLINE GEN
+vecsmall_reverse(GEN A)
+{
+  long i, l;
+  GEN B = cgetg_copy(A, &l);
+  for (i=1; i<l; i++) B[i] = A[l-i];
+  return B;
+}
+
+INLINE void
+vecreverse_inplace(GEN y)
+{
+  long ly = lg(y), lim = ly>>1, i;
+  for (i = 1; i <= lim; i++)
+  {
+    GEN z = gel(y,i);
+    gel(y,i)    = gel(y,ly-i);
+    gel(y,ly-i) = z;
+  }
+}
+
+INLINE GEN
+vecsmallpermute(GEN A, GEN p) { return perm_mul(A, p); }
+
 INLINE GEN
 vecpermute(GEN A, GEN p)
 {
@@ -884,7 +982,9 @@ rowpermute(GEN A, GEN p)
 {
   long i, lB = lg(A);
   GEN B = cgetg(lB, typ(A));
-  for (i=1; i<lB; i++) gel(B, i) = vecpermute(gel(A, i), p);
+  for (i=1; i<lB; i++)
+    gel(B, i) = typ(gel(A,i)) == t_VECSMALL ? vecsmallpermute(gel(A, i), p):
+                                                    vecpermute(gel(A, i), p);
   return B;
 }
 /*******************************************************************/
@@ -1089,7 +1189,7 @@ gerepilecopy(pari_sp av, GEN x)
   {
     avma = av;
     if (x < (GEN)av) {
-      if (x < (GEN)bot) new_chunk(lg(x));
+      if (x < (GEN)pari_mainstack->bot) new_chunk(lg(x));
       x = leafcopy_avma(x, av);
       avma = (pari_sp)x;
     } else
@@ -1133,6 +1233,7 @@ gerepileall(pari_sp av, int n, ...)
     for (--i; i>=0; i--) *gptr[i] = bin_copy((GENbin*)*gptr[i]);
     pari_free(gptr);
   }
+  va_end(a);
 }
 
 INLINE void
@@ -1156,10 +1257,7 @@ bin_copy(GENbin *p)
   base= p->base; dx = x - base;
   y = (GEN)memcpy((void*)new_chunk(len), (void*)GENbinbase(p), len*sizeof(long));
   y += dx;
-  if (p->canon)
-    shiftaddress_canon(y, (y-x)*sizeof(long));
-  else
-    shiftaddress(y, (y-x)*sizeof(long));
+  p->rebase(y, ((ulong)y-(ulong)x));
   pari_free(p); return y;
 }
 
@@ -1217,6 +1315,13 @@ gtos(GEN x) {
   return itos(x);
 }
 
+INLINE ulong
+gtou(GEN x) {
+  if (typ(x) != t_INT || signe(x)<0)
+    pari_err_TYPE("gtou [integer >=0 expected]",x);
+  return itou(x);
+}
+
 INLINE GEN
 absfrac(GEN x)
 {
@@ -1232,6 +1337,12 @@ Q_abs(GEN x) { return (typ(x) == t_INT)? absi(x): absfrac(x); }
 INLINE GEN
 Q_abs_shallow(GEN x)
 { return (typ(x) == t_INT)? absi_shallow(x): absfrac_shallow(x); }
+INLINE GEN
+R_abs_shallow(GEN x)
+{ return (typ(x) == t_FRAC)? absfrac_shallow(x): mpabs_shallow(x); }
+INLINE GEN
+R_abs(GEN x)
+{ return (typ(x) == t_FRAC)? absfrac(x): mpabs(x); }
 
 /* Force z to be of type real/complex with floating point components */
 INLINE GEN
@@ -1375,6 +1486,8 @@ INLINE long
 ndec2nlong(long x) { return 1 + (long)((x)*(LOG2_10/BITS_IN_LONG)); }
 INLINE long
 ndec2prec(long x) { return 2 + ndec2nlong(x); }
+INLINE long
+ndec2nbits(long x) { return ndec2nlong(x) << TWOPOTBITS_IN_LONG; }
 /* Fast implementation of ceil(x / (8*sizeof(long))); typecast to (ulong)
  * to avoid overflow. Faster than 1 + ((x-1)>>TWOPOTBITS_IN_LONG)) :
  *   addl, shrl instead of subl, sarl, addl */
@@ -1403,18 +1516,20 @@ nchar2nlong(long x) {
   return (long)(((ulong)x+sizeof(long)-1) >> (TWOPOTBITS_IN_LONG-3L));
 }
 INLINE long
-bit_accuracy(long x) { return (x-2) * BITS_IN_LONG; }
+prec2nbits(long x) { return (x-2) * BITS_IN_LONG; }
 INLINE double
 bit_accuracy_mul(long x, double y) { return (x-2) * (BITS_IN_LONG*y); }
 INLINE double
 prec2nbits_mul(long x, double y) { return (x-2) * (BITS_IN_LONG*y); }
 INLINE long
-bit_prec(GEN x) { return bit_accuracy(realprec(x)); }
+bit_prec(GEN x) { return prec2nbits(realprec(x)); }
 INLINE long
-prec2nbits(long x) { return bit_accuracy(x); }
+bit_accuracy(long x) { return prec2nbits(x); }
 INLINE long
 prec2ndec(long x) { return (long)prec2nbits_mul(x, LOG10_2); }
 INLINE long
+nbits2ndec(long x) { return (long)(x * LOG10_2); }
+INLINE long
 precdbl(long x) {return (x - 1) << 1;}
 INLINE long
 divsBIL(long n) { return n >> TWOPOTBITS_IN_LONG; }
@@ -1430,7 +1545,6 @@ remsBIL(long n) { return n & (BITS_IN_LONG-1); }
 
 INLINE GEN
 Fp_red(GEN a, GEN m) { return modii(a, m); }
-
 INLINE GEN
 Fp_add(GEN a, GEN b, GEN m)
 {
@@ -1488,10 +1602,18 @@ Fp_neg(GEN b, GEN m)
     p = remii(negi(b), m);
   return gerepileuptoint(av, p);
 }
+
+INLINE GEN
+Fp_halve(GEN a, GEN p)
+{
+  if (mpodd(a)) a = addii(a,p);
+  return shifti(a,-1);
+}
+
 /* assume 0 <= u < p and ps2 = p>>1 */
 INLINE GEN
 Fp_center(GEN u, GEN p, GEN ps2)
-{ return absi_cmp(u,ps2)<=0? icopy(u): subii(u,p); }
+{ return abscmpii(u,ps2)<=0? icopy(u): subii(u,p); }
 
 /* x + y*z mod p */
 INLINE GEN
@@ -1707,6 +1829,8 @@ is_noncalc_t(long tx) { return (tx) >= t_LIST; }
 INLINE int
 is_rational_t(long t) { return (t == t_INT || t == t_FRAC); }
 INLINE int
+is_real_t(long t) { return (t == t_INT || t == t_REAL || t == t_FRAC); }
+INLINE int
 is_recursive_t(long t) { return lontyp[t]; }
 INLINE int
 is_scalar_t(long t) { return (t < t_POL); }
@@ -1725,9 +1849,28 @@ sqrtr(GEN x) {
   if (s >= 0) return sqrtr_abs(x);
   retmkcomplex(gen_0, sqrtr_abs(x));
 }
+INLINE GEN
+cbrtr(GEN x) {
+  long s = signe(x);
+  GEN r;
+  if (s == 0) return real_0_bit(expo(x) / 3);
+  r = cbrtr_abs(x);
+  if (s < 0) togglesign(r);
+  return r;
+}
 /* x^(1/n) */
 INLINE GEN
-sqrtnr(GEN x, long n) { return mpexp(divrs(mplog(x), n)); }
+sqrtnr(GEN x, long n) {
+  switch(n)
+  {
+    case 1: return rcopy(x);
+    case 2: return sqrtr(x);
+    case 3: return cbrtr(x);
+  }
+  return mpexp(divrs(mplog(x), n));
+}
+INLINE long
+logint(GEN B, GEN y) { return logintall(B,y,NULL); }
 
 /*******************************************************************/
 /*                                                                 */
@@ -1752,11 +1895,9 @@ INLINE int odd(long x) { return x & 1; }
 INLINE int both_odd(long x, long y) { return x & y & 1; }
 
 INLINE int
-isonstack(GEN x) { return ((pari_sp)x >= bot && (pari_sp)x < top); }
-
-/* assume 0 <= k <= BITS_IN_LONG. Return uniform random 0 <= x < (1<<k) */
-INLINE long
-random_bits(long k) { return pari_rand() >> (BITS_IN_LONG - k); }
+isonstack(GEN x)
+{ return ((pari_sp)x >= pari_mainstack->bot
+       && (pari_sp)x <  pari_mainstack->top); }
 
 /* assume x != 0 and x t_REAL, return an approximation to log2(|x|) */
 INLINE double
@@ -1780,9 +1921,9 @@ mul_denom(GEN dx, GEN dy)
 
 /* POLYNOMIALS */
 INLINE GEN
-constant_term(GEN x) { return signe(x)? gel(x,2): gen_0; }
+constant_coeff(GEN x) { return signe(x)? gel(x,2): gen_0; }
 INLINE GEN
-leading_term(GEN x) { return lg(x) == 2? gen_0: gel(x,lg(x)-1); }
+leading_coeff(GEN x) { return lg(x) == 2? gen_0: gel(x,lg(x)-1); }
 INLINE ulong
 Flx_lead(GEN x) { return lg(x) == 2? 0: x[lg(x)-1]; }
 INLINE long
@@ -1814,6 +1955,14 @@ RgX_copy(GEN x)
   for (i = 2; i<lx; i++) gel(y,i) = gcopy(gel(x,i));
   return y;
 }
+/* have to use ulong to avoid silly warnings from gcc "assuming signed
+ * overflow does not occur" */
+INLINE GEN
+RgX_coeff(GEN x, long n)
+{
+  ulong l = lg(x);
+  return (n < 0 || ((ulong)n+3) > l)? gen_0: gel(x,n+2);
+}
 INLINE GEN
 RgX_renormalize(GEN x) { return RgX_renormalize_lg(x, lg(x)); }
 INLINE GEN
@@ -1859,6 +2008,8 @@ F2x_equal1(GEN x) { return Flx_equal1(x); }
 INLINE int
 F2x_equal(GEN V, GEN W) { return Flx_equal(V,W); }
 INLINE GEN
+F2x_copy(GEN x) { return leafcopy(x); }
+INLINE GEN
 F2v_copy(GEN x) { return leafcopy(x); }
 INLINE GEN
 Flv_copy(GEN x) { return leafcopy(x); }
@@ -1948,7 +2099,7 @@ zm_copy(GEN x) { return Flm_copy(x); }
 INLINE GEN
 zv_copy(GEN x) { return Flv_copy(x); }
 INLINE GEN
-row_zm(GEN x, long i) { return row_Flm(x,i); }
+zm_row(GEN x, long i) { return Flm_row(x,i); }
 
 INLINE GEN
 ZC_hnfrem(GEN x, GEN y) { return ZC_hnfremdiv(x,y,NULL); }
@@ -1989,14 +2140,25 @@ ZV_dvd(GEN x, GEN y)
 INLINE GEN
 Fq_red(GEN x, GEN T, GEN p)
 { return typ(x)==t_INT? Fp_red(x,p): FpXQ_red(x,T,p); }
+INLINE GEN
+Fq_to_FpXQ(GEN x, GEN T, GEN p /*unused*/)
+{
+  (void) p;
+  return typ(x)==t_INT ? scalarpol(x, get_FpX_var(T)): x;
+}
+
+INLINE GEN
+gener_Fq_local(GEN T, GEN p, GEN L)
+{ return T? gener_FpXQ_local(T,p, L)
+          : pgener_Fp_local(p, L); }
 
 /* FpXQX */
 INLINE GEN
 FpXQX_div(GEN x, GEN y, GEN T, GEN p) { return FpXQX_divrem(x, y, T, p, NULL); }
 INLINE GEN
-FlxqX_rem(GEN x, GEN y, GEN T, ulong p) { return FlxqX_divrem(x, y, T, p, ONLY_REM); }
-INLINE GEN
 FlxqX_div(GEN x, GEN y, GEN T, ulong p) { return FlxqX_divrem(x, y, T, p, NULL); }
+INLINE GEN
+F2xqX_div(GEN x, GEN y, GEN T) { return F2xqX_divrem(x, y, T, NULL); }
 
 /* FqX */
 INLINE GEN
@@ -2023,20 +2185,40 @@ INLINE GEN
 FqX_sqr(GEN x, GEN T, GEN p)
 { return T? FpXQX_sqr(x, T, p): FpX_sqr(x, p); }
 INLINE GEN
+FqX_powu(GEN x, ulong n, GEN T, GEN p)
+{ return T? FpXQX_powu(x, n, T, p): FpX_powu(x, n, p); }
+INLINE GEN
 FqX_div(GEN x, GEN y, GEN T, GEN p)
 { return T? FpXQX_divrem(x,y,T,p,NULL): FpX_divrem(x,y,p,NULL); }
 INLINE GEN
+FqX_get_red(GEN S, GEN T, GEN p)
+{ return T? FpXQX_get_red(S,T,p): FpX_get_red(S,p); }
+INLINE GEN
 FqX_rem(GEN x, GEN y, GEN T, GEN p)
 { return T? FpXQX_rem(x,y,T,p): FpX_rem(x,y,p); }
 INLINE GEN
 FqX_divrem(GEN x, GEN y, GEN T, GEN p, GEN *z)
 { return T? FpXQX_divrem(x,y,T,p,z): FpX_divrem(x,y,p,z); }
 INLINE GEN
+FqX_div_by_X_x(GEN x, GEN y, GEN T, GEN p, GEN *z)
+{ return T? FpXQX_div_by_X_x(x,y,T,p,z): FpX_div_by_X_x(x,y,p,z); }
+INLINE GEN
+FqX_halfgcd(GEN P,GEN Q,GEN T,GEN p)
+{return T? FpXQX_halfgcd(P,Q,T,p): FpX_halfgcd(P,Q,p);}
+INLINE GEN
 FqX_gcd(GEN P,GEN Q,GEN T,GEN p)
 {return T? FpXQX_gcd(P,Q,T,p): FpX_gcd(P,Q,p);}
 INLINE GEN
 FqX_extgcd(GEN P,GEN Q,GEN T,GEN p, GEN *U, GEN *V)
 { return T? FpXQX_extgcd(P,Q,T,p,U,V): FpX_extgcd(P,Q,p,U,V); }
+INLINE GEN
+FqX_normalize(GEN z, GEN T, GEN p) { return T? FpXQX_normalize(z, T, p): FpX_normalize(z, p); }
+INLINE GEN
+FqX_deriv(GEN f, /*unused*/GEN T, GEN p) { (void)T; return FpXX_deriv(f, p); }
+INLINE GEN
+FqX_factor(GEN f, GEN T, GEN p) { return T?FpXQX_factor(f, T, p): FpX_factor(f, p); }
+INLINE GEN
+FqX_roots(GEN f, GEN T, GEN p) { return T?FpXQX_roots(f, T, p): FpX_roots(f, p); }
 
 /*FqXQ*/
 INLINE GEN
@@ -2184,7 +2366,7 @@ quadnorm(GEN q)
     return gadd(z, gmul(c, gsqr(u)));
   }
 }
-/* x a t_QUAD, return the associated discriminant */
+/* x a t_QUAD, return the attached discriminant */
 INLINE GEN
 quad_disc(GEN x)
 {
@@ -2301,6 +2483,8 @@ INLINE long ellQp_get_prec(GEN E) { GEN z = ellQp_get_zero(E); return valp(z); }
 INLINE GEN ellQp_get_p(GEN E) { GEN z = ellQp_get_zero(E); return gel(z,2); }
 INLINE long ellR_get_prec(GEN x) { return nbits2prec(mael3(x, 15, 1, 1)); }
 INLINE long ellR_get_sign(GEN x) { return mael3(x, 15, 1, 2); }
+INLINE GEN ellnf_get_nf(GEN x) { return checknf_i(gmael(x,15,1)); }
+INLINE GEN ellnf_get_bnf(GEN x) { return checkbnf_i(gmael(x,15,1)); }
 
 INLINE int ell_is_inf(GEN z) { return lg(z) == 2; }
 INLINE GEN ellinf(void) { return mkvec(gen_0); }
@@ -2310,6 +2494,10 @@ INLINE GEN ellinf(void) { return mkvec(gen_0); }
 /*                    ALGEBRAIC NUMBER THEORY                      */
 /*                                                                 */
 /*******************************************************************/
+INLINE GEN modpr_get_pr(GEN x)  { return gel(x,3); }
+INLINE GEN modpr_get_p(GEN x)  { return pr_get_p(modpr_get_pr(x)); }
+INLINE GEN modpr_get_T(GEN x)  { return lg(x) == 4? NULL: gel(x,4); }
+
 INLINE GEN pr_get_p(GEN pr)  { return gel(pr,1); }
 INLINE GEN pr_get_gen(GEN pr){ return gel(pr,2); }
 /* .[2] instead of itos works: e and f are small positive integers */
@@ -2320,6 +2508,8 @@ INLINE int
 pr_is_inert(GEN P) { return pr_get_f(P) == lg(pr_get_gen(P))-1; }
 INLINE GEN
 pr_norm(GEN pr) { return powiu(pr_get_p(pr), pr_get_f(pr)); }
+INLINE ulong
+upr_norm(GEN pr) { return upowuu(pr_get_p(pr)[2], pr_get_f(pr)); }
 
 /* assume nf a genuine nf */
 INLINE long
@@ -2347,6 +2537,8 @@ nf_get_Tr(GEN nf) { return gmael(nf,5,4); }
 INLINE GEN
 nf_get_diff(GEN nf) { return gmael(nf,5,5); }
 INLINE GEN
+nf_get_ramified_primes(GEN nf) { return gmael(nf,5,8); }
+INLINE GEN
 nf_get_roots(GEN nf) { return gel(nf,6); }
 INLINE GEN
 nf_get_zk(GEN nf) { return gel(nf,7); }
@@ -2426,6 +2618,18 @@ bid_get_arch(GEN bid) { return gmael(bid,1,2); }
 INLINE GEN
 bid_get_grp(GEN bid) { return gel(bid,2); }
 INLINE GEN
+bid_get_fact(GEN bid) { return gel(bid,3); }
+INLINE GEN
+bid_get_sprk(GEN bid) { return gmael(bid,4,1); }
+INLINE GEN
+bid_get_sarch(GEN bid) { return gmael(bid,4,2); }
+INLINE GEN
+bid_get_archp(GEN bid) { return gmael3(bid,4,2,4); }
+INLINE GEN
+bid_get_ind(GEN bid) { return gmael(bid,4,3); }
+INLINE GEN
+bid_get_U(GEN bid) { return gel(bid,5); }
+INLINE GEN
 bid_get_no(GEN bid) { return abgrp_get_no(bid_get_grp(bid)); }
 INLINE GEN
 bid_get_cyc(GEN bid) { return abgrp_get_cyc(bid_get_grp(bid)); }
@@ -2466,6 +2670,12 @@ rnf_get_nfdegree(GEN rnf) { return degpol(nf_get_pol(rnf_get_nf(rnf))); }
 INLINE long
 rnf_get_absdegree(GEN rnf) { return degpol(gmael(rnf,11,1)); }
 INLINE GEN
+rnf_get_idealdisc(GEN rnf) { return gmael(rnf,3,1); }
+INLINE GEN
+rnf_get_k(GEN rnf) { return gmael(rnf,11,3); }
+INLINE GEN
+rnf_get_alpha(GEN rnf) { return gmael(rnf, 11, 2); }
+INLINE GEN
 rnf_get_nf(GEN rnf) { return gel(rnf,10); }
 INLINE void
 rnf_get_nfzk(GEN rnf, GEN *b, GEN *cb) {*b=gmael(rnf,2,1); *cb=gmael(rnf,2,2);}
@@ -2490,6 +2700,12 @@ rnf_get_map(GEN rnf) { return gel(rnf,11); }
 INLINE GEN
 rnf_get_invzk(GEN rnf) { return gel(rnf,8); }
 
+/* I integral ZM (not HNF), G ZM, rounded Cholesky form of a weighted
+ * T2 matrix. Reduce I wrt G */
+INLINE GEN
+idealpseudored(GEN I, GEN G)
+{ return ZM_mul(I, ZM_lll(ZM_mul(G, I), 0.99, LLL_IM)); }
+
 /* I integral (not necessarily HNF), G ZM, rounded Cholesky form of a weighted
  * T2 matrix. Return m in I with T2(m) small */
 INLINE GEN
@@ -2517,12 +2733,17 @@ idealred_elt(GEN nf, GEN I) {
 INLINE GEN
 idealred(GEN nf, GEN I) { return idealred0(nf, I, NULL); }
 
+INLINE GEN
+idealchineseinit(GEN nf, GEN x)
+{ return idealchinese(nf,x,NULL); }
+
 /*******************************************************************/
 /*                                                                 */
 /*                              CLOSURES                           */
 /*                                                                 */
 /*******************************************************************/
-INLINE long closure_arity(GEN C)    { return C[1]; }
+INLINE long closure_arity(GEN C)          { return ((ulong)C[1])&ARITYBITS; }
+INLINE long closure_is_variadic(GEN C) { return !!(((ulong)C[1])&VARARGBITS); }
 INLINE const char *closure_codestr(GEN C)  { return GSTR(gel(C,2))-1; }
 INLINE GEN closure_get_code(GEN C)  { return gel(C,2); }
 INLINE GEN closure_get_oper(GEN C)  { return gel(C,3); }
