@@ -24,7 +24,7 @@ static const ulong hashprimes[] = {
   393241, 786433, 1572869, 3145739, 6291469, 12582917, 25165843, 50331653,
   100663319, 201326611, 402653189, 805306457, 1610612741
 };
-static const int hashprimes_len = sizeof(hashprimes) / sizeof(hashprimes[0]);
+static const int hashprimes_len = numberof(hashprimes);
 
 INLINE void
 setlen(hashtable *h, ulong len) {
@@ -44,12 +44,14 @@ get_prime_index(ulong len)
 
 /* link hashentry e to hashtable h, setting e->hash / e->next */
 INLINE void
-hash_link(hashtable *h, hashentry *e)
+hash_link2(hashtable *h, hashentry *e, ulong hash)
 {
   ulong index;
-  e->hash = h->hash(e->key); index = e->hash % h->len;
+  e->hash = hash; index = e->hash % h->len;
   e->next = h->table[index]; h->table[index] = e;
 }
+INLINE void
+hash_link(hashtable *h, hashentry *e) { hash_link2(h,e,h->hash(e->key));}
 
 hashtable *
 hash_create(ulong minsize, ulong (*hash)(void*), int (*eq)(void*,void*),
@@ -79,7 +81,7 @@ hash_create(ulong minsize, ulong (*hash)(void*), int (*eq)(void*,void*),
 }
 
 void
-hash_insert(hashtable *h, void *k, void *v)
+hash_insert2(hashtable *h, void *k, void *v, ulong hash)
 {
   hashentry *e;
   ulong index;
@@ -110,18 +112,59 @@ hash_insert(hashtable *h, void *k, void *v)
     setlen(h, newlen);
   }
   e->key = k;
-  e->val = v; hash_link(h, e);
+  e->val = v; hash_link2(h, e, hash);
+}
+void
+hash_insert(hashtable *h, void *k, void *v)
+{ hash_insert2(h,k,v,h->hash(k)); }
+
+/* the key 'k' may correspond to different values in the hash, return
+ * one satisfying the selection callback */
+hashentry *
+hash_select(hashtable *h, void *k, void *E,int(*select)(void *,hashentry *))
+{
+  ulong hash = h->hash(k);
+  hashentry *e = h->table[ hash % h->len ];
+  while (e)
+  {
+    if (hash == e->hash && h->eq(k, e->key) && select(E,e)) return e;
+    e = e->next;
+  }
+  return NULL;
 }
 
-/* returns entry associated with key k or NULL */
-hashentry *
-hash_search(hashtable *h, void *k)
+GEN
+hash_keys(hashtable *h)
 {
-  ulong hash;
-  hashentry *e;
-  if (h->nb == 0) return NULL;
-  hash = h->hash(k);
-  e = h->table[ hash % h->len ];
+  long k = 1;
+  ulong i;
+  GEN v = cgetg(h->nb+1, t_VECSMALL);
+  for (i = 0; i < h->len; i++)
+  {
+    hashentry *e = h->table[i];
+    while (e) { v[k++] = (long)e->key; e = e->next; }
+  }
+  return v;
+}
+GEN
+hash_values(hashtable *h)
+{
+  long k = 1;
+  ulong i;
+  GEN v = cgetg(h->nb+1, t_VECSMALL);
+  for (i = 0; i < h->len; i++)
+  {
+    hashentry *e = h->table[i];
+    while (e) { v[k++] = (long)e->val; e = e->next; }
+  }
+  return v;
+}
+
+/* assume hash = h->hash(k) */
+hashentry *
+hash_search2(hashtable *h, void *k, ulong hash)
+{
+  hashentry *e = h->table[ hash % h->len ];
   while (e)
   {
     if (hash == e->hash && h->eq(k, e->key)) return e;
@@ -129,6 +172,31 @@ hash_search(hashtable *h, void *k)
   }
   return NULL; /* not found */
 }
+/* returns entry attached to key k or NULL */
+hashentry *
+hash_search(hashtable *h, void *k)
+{
+  if (h->nb == 0) return NULL;
+  return hash_search2(h, k, h->hash(k));
+}
+
+hashentry *
+hash_remove_select(hashtable *h, void *k, void *E,
+  int (*select)(void*,hashentry*))
+{
+  ulong hash = h->hash(k), index = hash % h->len;
+  hashentry **pE = &(h->table[index]), *e = *pE;
+  while (e)
+  {
+    if (hash == e->hash && h->eq(k, e->key) && select(E,e)) {
+      *pE = e->next; h->nb--;
+      return e;
+    }
+    pE = &(e->next);
+    e = e->next;
+  }
+  return NULL;
+}
 
 hashentry *
 hash_remove(hashtable *h, void *k)
@@ -158,19 +226,30 @@ hash_destroy(hashtable *h)
   }
   pari_free(h->table); pari_free(h);
 }
+static ulong
+hash_id(void *x) { return (ulong)x; }
+static int
+eq_id(void *x, void *y) { return x == y; }
+hashtable *
+hash_create_ulong(ulong s, long stack)
+{ return hash_create(s, &hash_id, &eq_id, stack); }
 
 static
 int strequal(void *a, void *b) { return !strcmp((char*)a,(char*)b); }
+hashtable *
+hash_create_str(ulong s, long stack)
+{ return hash_create(s, (ulong (*)(void *))&hash_str, strequal, stack); }
+
 hashtable *
 hashstr_import_static(hashentry *e, ulong size)
 {
-  hashtable *h = hash_create(size, (ulong (*)(void *))hash_str, strequal, 0);
+  hashtable *h = hash_create_str(size, 0);
   for ( ; e->key; e++) { hash_link(h, e); h->nb++; }
   return h;
 }
 
 void
-hashstr_dbg(hashtable *h)
+hash_dbg(hashtable *h)
 {
   ulong n, Total = 0, Max = 0;
   hashentry *e, **table = h->table;
@@ -192,7 +271,7 @@ hashstr_dbg(hashtable *h)
 /********************************************************************/
 
 INLINE ulong
-glue(ulong h, ulong a) { return (h << 5) + (h>>2) + a; }
+glue(ulong h, ulong a) { return 404936533*h + a; }
 ulong
 hash_GEN(GEN x)
 {
@@ -203,13 +282,13 @@ hash_GEN(GEN x)
     case t_INT:
       lx = lgefint(x);
       h &= TYPBITS;
-      for (i = 1; i < lx; i++) h = glue(h, (ulong)x[i]);
+      for (i = 1; i < lx; i++) h = glue(h, uel(x,i));
       return h;
     case t_REAL:
     case t_STR:
     case t_VECSMALL:
       lx = lg(x);
-      for (i = 1; i < lx; i++) h = glue(h, (ulong)x[i]);
+      for (i = 1; i < lx; i++) h = glue(h, uel(x,i));
       return h;
     /* one more special case */
     case t_LIST:
