@@ -13,6 +13,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 /* Originally written by Vasili Burdo */
 
+#include "pwinver.h"
 #include <windows.h>
 #include <stdio.h>
 #include "mingw.h"
@@ -156,3 +157,58 @@ win32_set_pdf_viewer(void)
     putenv(buf);
   }
 }
+
+extern int win32ctrlc, win32alrm;
+static HANDLE hTimerQueue = NULL;
+
+static void CALLBACK
+win32_cb_alarm(void *lpParam, BOOLEAN TimerOrWaitFired)
+{
+  (void) lpParam; (void) TimerOrWaitFired;
+  win32ctrlc++;
+  win32alrm = 1;
+}
+
+void
+win32_alarm(unsigned int s)
+{
+  if (hTimerQueue)
+  {
+    HANDLE oldhTimerQueue = hTimerQueue;
+    hTimerQueue = NULL;
+    DeleteTimerQueue(oldhTimerQueue);
+  }
+  if (s)
+  {
+    void *arg = NULL;
+    HANDLE hTimer = NULL;
+    hTimerQueue = CreateTimerQueue();
+    CreateTimerQueueTimer( &hTimer, hTimerQueue,
+        (WAITORTIMERCALLBACK)win32_cb_alarm, &arg , s*1000, 0, 0);
+  }
+}
+
+#define WIN32_FILETIME_PER_MILLISECOND  10000
+
+long
+win32_timer(void)
+{
+  FILETIME lpCreation, lpExit, lpKernel, lpUser;
+  LARGE_INTEGER time;
+  GetProcessTimes(
+    GetCurrentProcess(),
+    &lpCreation, &lpExit, &lpKernel, &lpUser
+    );
+  time.HighPart = lpUser.dwHighDateTime;
+  time.LowPart = lpUser.dwLowDateTime;
+  time.QuadPart /= WIN32_FILETIME_PER_MILLISECOND;
+  return time.LowPart;
+}
+
+long
+win32_nbthreads(void)
+{
+  SYSTEM_INFO system_info;
+  GetSystemInfo(&system_info);
+  return system_info.dwNumberOfProcessors;
+}
